######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains the :class:`MappingListModel` model."""
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt, Signal
from spinedb_api.export_mapping.export_mapping import (
    DimensionMapping,
    EntityClassMapping,
    ParameterDefaultValueIndexMapping,
    ParameterValueIndexMapping,
)
from spinetoolbox.helpers import unique_name


class MappingsTableModel(QAbstractTableModel):
    """A table model that holds export mappings."""

    rename_requested = Signal(int, str)
    """Emitted when mapping's name should be changed."""
    mapping_enabled_state_change_requested = Signal(int)
    """Emitted when mapping's enabled state should be changed."""
    set_all_mappings_enabled_requested = Signal(bool)
    """Emitted when all mappings should be enabled/disabled."""
    write_order_about_to_change = Signal()
    """Emitted before changing the write order."""
    write_order_changed = Signal()
    """Emitted after the write order has changed."""

    MAPPING_SPECIFICATION_ROLE = Qt.ItemDataRole.UserRole + 1
    MAPPING_TYPE_ROLE = Qt.ItemDataRole.UserRole + 2
    MAPPING_ROOT_ROLE = Qt.ItemDataRole.UserRole + 3
    ALWAYS_EXPORT_HEADER_ROLE = Qt.ItemDataRole.UserRole + 4
    ENTITY_DIMENSIONS_ROLE = Qt.ItemDataRole.UserRole + 5
    USE_FIXED_TABLE_NAME_FLAG_ROLE = Qt.ItemDataRole.UserRole + 6
    FIXED_TABLE_NAME_ROLE = Qt.ItemDataRole.UserRole + 7
    PARAMETER_DIMENSIONS_ROLE = Qt.ItemDataRole.UserRole + 8
    GROUP_FN_ROLE = Qt.ItemDataRole.UserRole + 9
    HIGHLIGHT_POSITION_ROLE = Qt.ItemDataRole.UserRole + 10

    def __init__(self, mappings=None, parent=None):
        """
        Args:
            mappings (dict, optional): mapping from name to ``MappingSpecification``
            parent (QObject, optional): parent object
        """
        super().__init__(parent)
        if mappings is None:
            mappings = {}
        self._names = list(mappings)
        self._mappings = mappings

    def columnCount(self, parent=QModelIndex()):
        return 2

    def extend(self, mapping_specification, name=""):
        """
        Appends a mapping to the table.

        Args:
            mapping_specification (MappingSpecification): specification to add
            name (str): mapping specification's name; if empty, the mapping is given a default name

        Returns:
            int: row index of the new mapping
        """
        position = len(self._names)
        self.beginInsertRows(QModelIndex(), position, position)
        if not name:
            name = unique_name("Mapping", self._names)
        elif name in self._names:
            name = unique_name(name, self._names)
        self._mappings[name] = mapping_specification
        self._names.append(name)
        self.endInsertRows()
        return position

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        column = index.column()
        if column == 0:
            if role == Qt.ItemDataRole.CheckStateRole:
                spec = self._mappings[self._names[index.row()]]
                return Qt.CheckState.Checked if spec.enabled else Qt.CheckState.Unchecked
            if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
                return self._names[index.row()]
        if column == 1:
            if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
                order = str(index.row() + 1)
                if order.endswith(("11", "12", "13")):
                    return order + "th"
                if order.endswith("1"):
                    return order + "st"
                if order.endswith("2"):
                    return order + "nd"
                if order.endswith("3"):
                    return order + "rd"
                return order + "th"
            if role == Qt.TextAlignmentRole:
                return Qt.AlignCenter
        if role >= Qt.ItemDataRole.UserRole:
            spec = self._mappings[self._names[index.row()]]
            if role == self.MAPPING_SPECIFICATION_ROLE:
                return spec
            if role == self.MAPPING_TYPE_ROLE:
                return spec.type.value
            if role == self.MAPPING_ROOT_ROLE:
                return spec.root
            if role == self.ALWAYS_EXPORT_HEADER_ROLE:
                return spec.always_export_header
            if role == self.ENTITY_DIMENSIONS_ROLE:
                return _instance_occurrences(spec.root, DimensionMapping)
            if role == self.USE_FIXED_TABLE_NAME_FLAG_ROLE:
                return spec.use_fixed_table_name_flag
            if role == self.FIXED_TABLE_NAME_ROLE:
                return spec.root.value
            if role == self.PARAMETER_DIMENSIONS_ROLE:
                dimensions = _instance_occurrences(spec.root, ParameterValueIndexMapping)
                if dimensions == 0:
                    return _instance_occurrences(spec.root, ParameterDefaultValueIndexMapping)
                return dimensions
            if role == self.GROUP_FN_ROLE:
                return spec.group_fn
            if role == self.HIGHLIGHT_POSITION_ROLE:
                highlighting_mapping = next((m for m in spec.root.flatten() if isinstance(m, EntityClassMapping)), None)
                if highlighting_mapping is None:
                    return None
                return highlighting_mapping.highlight_position
        return None

    def flags(self, index):
        if index.column() == 0:
            return super().flags(index) | Qt.ItemIsUserCheckable | Qt.ItemIsEditable
        return super().flags(index)

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return ("Name", "Write order")[section]
        return None

    def mapping_specification(self, name):
        """
        Returns a mapping specification for given name.

        Args:
            name (str): mapping's name

        Returns:
            MappingSpecification: mapping specification
        """
        return self._mappings[name]

    def index_of(self, name):
        """
        Returns index for given mapping's name.

        Args:
            name (str): mapping's name

        Returns:
            QModelIndex: index
        """
        try:
            row = self._names.index(name)
        except ValueError:
            return QModelIndex()
        return self.index(row, 0)

    def reorder_writing(self, row, earlier):
        """Reorders writings.

        Args:
            row (int): row to reorder
            earlier (bool): True to write earlier, False to write later
        """
        moved = [self._names[row]]
        if earlier:
            before = self._names[: row - 1]
            after = [self._names[row - 1]] + self._names[row + 1 :]
        else:
            before = self._names[:row] + [self._names[row + 1]]
            after = self._names[row + 2 :]
        self.write_order_about_to_change.emit()
        self.beginRemoveRows(QModelIndex(), row, row)
        self._names = before + after
        self.endRemoveRows()
        insert_row = row - 1 if earlier else row + 1
        self.beginInsertRows(QModelIndex(), insert_row, insert_row)
        self._names = before + moved + after
        self.endInsertRows()
        reordered_mappings = {name: self._mappings[name] for name in self._names}
        self._mappings.clear()
        self._mappings.update(reordered_mappings)
        self.write_order_changed.emit()

    def rowCount(self, parent=QModelIndex()):
        return len(self._names)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        column = index.column()
        if column == 0:
            if role == Qt.ItemDataRole.CheckStateRole:
                row = index.row()
                self.mapping_enabled_state_change_requested.emit(row)
                return True
            if role == Qt.ItemDataRole.EditRole:
                if not value or value in self._names:
                    return False
                self.rename_requested.emit(index.row(), value)
                return True
        if role > Qt.ItemDataRole.UserRole:
            name = self._names[index.row()]
            spec = self._mappings[name]
            if role == self.MAPPING_ROOT_ROLE:
                spec.root = value
                self.dataChanged.emit(index, index, [self.MAPPING_ROOT_ROLE])
            elif role == self.MAPPING_TYPE_ROLE:
                spec.type = value
                self.dataChanged.emit(index, index, [self.MAPPING_TYPE_ROLE])
            elif role == self.ALWAYS_EXPORT_HEADER_ROLE:
                spec.always_export_header = value
                self.dataChanged.emit(index, index, [self.ALWAYS_EXPORT_HEADER_ROLE])
            elif role == self.USE_FIXED_TABLE_NAME_FLAG_ROLE:
                spec.use_fixed_table_name_flag = value
                self.dataChanged.emit(index, index, [self.USE_FIXED_TABLE_NAME_FLAG_ROLE])
            elif role == self.FIXED_TABLE_NAME_ROLE:
                spec.root.value = value
                self.dataChanged.emit(index, index, [self.FIXED_TABLE_NAME_ROLE])
            elif role == self.GROUP_FN_ROLE:
                spec.group_fn = value
                self.dataChanged.emit(index, index, [self.GROUP_FN_ROLE])
            elif role == self.HIGHLIGHT_POSITION_ROLE:
                highlighting_mapping = next((m for m in spec.root.flatten() if isinstance(m, EntityClassMapping)), None)
                if highlighting_mapping is None:
                    return False
                highlighting_mapping.highlight_position = value
                self.dataChanged.emit(index, index, [self.HIGHLIGHT_POSITION_ROLE])
            return True
        return False

    def insert_mapping(self, row, name, mapping_specification):
        """
        Adds a new mapping.

        Args:
            row (int): row index
            name (str): mapping's name
            mapping_specification (MappingSpecification): mapping specification
        """
        self.beginInsertRows(QModelIndex(), row, row)
        self._names.insert(row, name)
        self._mappings[name] = mapping_specification
        self.endInsertRows()

    def remove_mapping(self, name):
        """
        Deletes a mapping.

        Args:
            name (str): mapping's name
        """
        row = self._names.index(name)
        self.beginRemoveRows(QModelIndex(), row, row)
        name = self._names.pop(row)
        del self._mappings[name]
        self.endRemoveRows()

    def rename_mapping(self, row, new_name):
        """
        Renames an item.

        Args:
            row (int): row index
            new_name (str): item's new name
        """
        previous = self._names[row]
        self._names[row] = new_name
        self._mappings[new_name] = self._mappings.pop(previous)
        index = self.index(row, 1)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])

    def set_mapping_enabled(self, row, enabled):
        """Enables or disables a mapping.

        Args:
            row (int): row index
            enabled (bool): True to enable mapping, False to disable
        """
        name = self._names[row]
        spec = self._mappings[name]
        spec.enabled = enabled
        index = self.index(row, 0)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.CheckStateRole])

    def set_all_enabled(self, enabled):
        """Enables/disables all mappings.

        Args:
            enabled (bool): True to enable, False to disable
        """
        first = None
        last = 0
        for i, name in enumerate(self._names):
            mapping = self._mappings[name]
            if mapping.enabled != enabled:
                mapping.enabled = enabled
                if first is None:
                    first = i
                last = i
        if first is not None:
            top_left = self.index(first, 0)
            bottom_right = self.index(last, 0)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.CheckStateRole])

    def enable_mapping_rows(self, rows):
        """Enables mappings on given rows and disables the rest.

        Args:
            rows (set of int): row indexes to enable
        """
        first = None
        last = 0
        for row, name in enumerate(self._names):
            mapping = self._mappings[name]
            enabled = row in rows
            if mapping.enabled != enabled:
                mapping.enabled = enabled
                if first is None:
                    first = row
                last = row
        if first is not None:
            top_left = self.index(first, 0)
            bottom_right = self.index(last, 0)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.CheckStateRole])

    def enabled_mapping_rows(self):
        """Returns enabled mapping rows.

        Return:
            set of int: enabled mapping row indexes.
        """
        rows = set()
        for row, name in enumerate(self._names):
            if self._mappings[name].enabled:
                rows.add(row)
        return rows

    def reset(self, mappings):
        self.beginResetModel()
        self._names = list(mappings)
        self._mappings = mappings
        self.endResetModel()


def _has_mapping_instance(mapping, cls):
    """
    Checks if mapping or any child mappings is an instance of ``cls``.

    Args:
        mapping (Mapping): a mapping
        cls (Type): subclass of ``Mapping`` to check against

    Returns:
        bool: True if mapping or any child mappings is an instance of ``cls``, False otherwise
    """
    if isinstance(mapping, cls):
        return True
    if mapping.child is not None:
        return _has_mapping_instance(mapping.child, cls)
    return False


def _instance_occurrences(mapping, cls):
    """
    Counts occurrences of ``cls`` in mapping hierarchy.

    Args:
        mapping (Mapping): a mapping
        cls (Type): type to count

    Returns:
        int: number of instances of ``cls``
    """
    count = 0
    if isinstance(mapping, cls):
        count += 1
    if mapping.child is not None:
        count += _instance_occurrences(mapping.child, cls)
    return count



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Exporter's executable item as well as support utilities."""
import json
from json import dump
import os
from pathlib import Path
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spine_engine.project_item.project_item_resource import file_resource_in_pack
from spine_engine.spine_engine import ItemExecutionFinishState
from spine_engine.utils.returning_process import ReturningProcess
from spine_engine.utils.serialization import deserialize_path
from spinedb_api import clear_filter_configs
from spinedb_api.spine_io import gdx_utils
from ..utils import generate_filter_subdirectory_name
from .do_work import do_work
from .item_info import ItemInfo
from .output_channel import OutputChannel
from .specification import OutputFormat
from .utils import EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX, Database, output_database_resources


class ExecutableItem(ExecutableItemBase):
    def __init__(
        self, name, specification, output_channels, output_time_stamps, cancel_on_error, gams_path, project_dir, logger
    ):
        """
        Args:
            name (str): item's name
            specification (ExporterSpecification): export settings
            output_channels (list of OutputChannel): output labels
            output_time_stamps (bool): if True append output directories with time stamps
            cancel_on_error (bool): if True execution fails on all errors else some errors can be ignored
            gams_path (str): GAMS path from Toolbox settings
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        self._output_time_stamps = output_time_stamps
        self._cancel_on_error = cancel_on_error
        self._gams_path = gams_path
        self._forks = {}
        self._result_files = {}
        self._process = None
        self._specification = specification
        self._output_channels = output_channels

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def _database_out_labels(self, resources):
        """
        Connects full database urls to output labels.

        Args:
            resources (Iterable of ProjectItemResource): forward database resources

        Returns:
            dict: a mapping from full database URL to output label
        """
        databases = {}
        for resource in resources:
            channel = next((channel for channel in self._output_channels if channel.in_label == resource.label), None)
            if channel is None:
                channel = self._find_legacy_output_channel(resource)
                if channel is None:
                    self._logger.msg_warning.emit(
                        f"<b>{self.name}</b>: No settings for database {resource.label}. Skipping."
                    )
                    continue
            if not channel.out_label:
                self._logger.msg_warning.emit(
                    f"<b>{self.name}</b>: No output label given to database {resource.label}. Skipping."
                )
                continue
            databases[resource.url] = channel.out_label
        return databases

    def _find_legacy_output_channel(self, resource):
        """
        Find legacy output channel that corresponds to given resource

        Args:
            resource (ProjectItemResource): database resource

        Returns:
            OutputChannel: output channel or None if not found
        """
        in_label = clear_filter_configs(resource.url)
        channel = next((c for c in self._output_channels if c.in_label == in_label), None)
        if channel is not None:
            channel.in_label = resource.label
        return channel

    def _out_urls(self, resources):
        """Returns output URLs for given input databases.

        Args:
            resources (Iterable of ProjectItemResource): forward database resources

        Returns:
            dict: a mapping from full database URL to output URL
        """
        out_urls = {}
        for resource in resources:
            channel = next((channel for channel in self._output_channels if channel.in_label == resource.label), None)
            if channel is None:
                channel = self._find_legacy_output_channel(resource)
                if channel is None:
                    continue
            if not channel.out_label:
                continue
            out_urls[resource.url] = channel.out_url
        return out_urls

    def stop_execution(self):
        """Stops executing this item."""
        super().stop_execution()
        if self._process is not None:
            self._process.terminate()
            self._process = None

    def execute(self, forward_resources, backward_resources, lock):
        """See base class."""
        status = super().execute(forward_resources, backward_resources, lock)
        if status != ItemExecutionFinishState.SUCCESS:
            return status
        if self._specification is None:
            self._logger.msg_warning.emit(f"<b>{self.name}</b>: No export settings configured. Skipping.")
            return ItemExecutionFinishState.SKIPPED
        database_resources = tuple(r for r in forward_resources if r.type_ == "database")
        databases = self._database_out_labels(database_resources)
        if not databases:
            return ItemExecutionFinishState.SKIPPED
        out_urls = self._out_urls(database_resources)
        gams_system_directory = ""
        if self._specification.output_format == OutputFormat.GDX:
            gams_system_directory = self._resolve_gams_system_directory()
            if gams_system_directory is None:
                self._logger.msg_error.emit(f"<b>{self.name}</b>: Cannot proceed. No GAMS installation found.")
                return ItemExecutionFinishState.FAILURE
        out_dir = Path(self._data_dir, "output")
        self._process = ReturningProcess(
            target=do_work,
            args=(
                self._specification.to_dict(),
                self._output_time_stamps,
                self._cancel_on_error,
                gams_system_directory,
                str(out_dir),
                databases,
                out_urls,
                self._filter_id,
                generate_filter_subdirectory_name(forward_resources, self.hash_filter_id()),
                self._logger,
            ),
        )
        result = self._process.run_until_complete()
        if len(result) > 1:
            self._result_files = result[1]
            file_name = self._manifest_file_name()
            with open(Path(self._data_dir, file_name), "w") as manifest:
                dump(
                    {
                        label: list(str(Path(file).relative_to(Path(self._data_dir))) for file in files)
                        for label, files in self._result_files.items()
                    },
                    manifest,
                )
        self._process = None
        return ItemExecutionFinishState.SUCCESS if result[0] else ItemExecutionFinishState.FAILURE

    def exclude_execution(self, forward_resources, backward_resources, lock):
        """See base class."""
        manifest_file_name = self._manifest_file_name()
        manifest_file_path = Path(self._data_dir, manifest_file_name)
        if not manifest_file_path.is_file():
            return
        with open(Path(self._data_dir, manifest_file_name)) as manifest_file:
            try:
                manifest = json.load(manifest_file)
            except json.decoder.JSONDecodeError:
                return
        self._result_files = {label: set(files) for label, files in manifest.items()}

    def _output_resources_forward(self):
        """See base class."""
        resources = []
        for label, output_files in self._result_files.items():
            resources += [file_resource_in_pack(self.name, label, str(Path(self._data_dir, f))) for f in output_files]
        if self._specification is not None and self._specification.output_format == OutputFormat.SQL:
            resources += output_database_resources(self.name, self._output_channels)
        return resources

    def _manifest_file_name(self):
        """Creates file name for manifest file.

        Returns:
            str: file name
        """
        return (
            EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX + (f"-{self.hash_filter_id()}" if self._filter_id else "")
        ) + ".json"

    def _resolve_gams_system_directory(self):
        """Returns GAMS system path from Toolbox settings or None if GAMS default is to be used.

        Returns:
            str: GAMS system path
        """
        path = self._gams_path
        if not path:
            path = gdx_utils.find_gams_directory()
        if path is not None and os.path.isfile(path):
            path = os.path.dirname(path)
        return path

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        specification_name = item_dict["specification"]
        specification = cls._get_specification(name, ItemInfo.item_type(), specification_name, specifications, logger)
        output_channels = [OutputChannel.from_dict(d, name, project_dir) for d in item_dict.get("output_labels", [])]
        for db_dict in item_dict.get("databases", []):
            # legacy
            db = Database.from_dict(db_dict)
            db.url = clear_filter_configs(deserialize_path(db_dict["database_url"], project_dir))
            output_channels.append(OutputChannel(db.url, name, db.output_file_name))
        for in_label, credentials in item_dict.get("db_credentials", {}).items():
            for channel in output_channels:
                if channel.in_label == in_label:
                    channel.out_url.update(credentials)
        output_time_stamps = item_dict.get("output_time_stamps", False)
        cancel_on_error = item_dict.get("cancel_on_error", True)
        gams_path = app_settings.value("appSettings/gamsPath", defaultValue=None)
        return ExecutableItem(
            name, specification, output_channels, output_time_stamps, cancel_on_error, gams_path, project_dir, logger
        )



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ``database_validation`` module."""
import gc
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from PySide6.QtCore import Slot
from PySide6.QtWidgets import QApplication
from sqlalchemy.engine.url import make_url
from spine_items.database_validation import DatabaseConnectionValidator
from spinedb_api import create_new_spine_database


class TestDatabaseConnectionValidator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_successful_validation_of_sqlite_database(self):
        with TemporaryDirectory() as temp_dir:
            url = "sqlite:///" + str(Path(temp_dir, "db.sqlite"))
            create_new_spine_database(url)
            listener = _Listener()
            validator = DatabaseConnectionValidator()
            try:
                sa_url = make_url(url)
                validator.validate_url("sqlite", sa_url, listener.failure, listener.success)
                while not listener.is_done:
                    QApplication.processEvents()
            finally:
                validator.wait_for_finish()
                validator.deleteLater()
            self.assertTrue(listener.is_success)
            gc.collect()

    def test_successful_validation_of_sqlite_database_with_str_url(self):
        with TemporaryDirectory() as temp_dir:
            url = "sqlite:///" + str(Path(temp_dir, "db.sqlite"))
            create_new_spine_database(url)
            listener = _Listener()
            validator = DatabaseConnectionValidator()
            try:
                sa_url = make_url(url)
                validator.validate_url("sqlite", sa_url, listener.failure, listener.success)
                while not listener.is_done:
                    QApplication.processEvents()
            finally:
                validator.wait_for_finish()
                validator.deleteLater()
            self.assertTrue(listener.is_success)
            gc.collect()

    def test_validation_failure_due_to_missing_sqlite_file(self):
        with TemporaryDirectory() as temp_dir:
            url = "sqlite:///" + str(Path(temp_dir, "db.sqlite"))
            listener = _Listener()
            validator = DatabaseConnectionValidator()
            try:
                sa_url = make_url(url)
                validator.validate_url("sqlite", sa_url, listener.failure, listener.success)
                while not listener.is_done:
                    QApplication.processEvents()
            finally:
                validator.wait_for_finish()
                validator.deleteLater()
            self.assertFalse(listener.is_success)
            self.assertEqual(listener.fail_message, "File does not exist. Check the Database field in the URL.")
            gc.collect()


class _Listener:
    def __init__(self):
        self.fail_message = None
        self._is_done = False

    @property
    def is_done(self):
        return self._is_done

    @property
    def is_success(self):
        if not self._is_done:
            raise RuntimeError("No success status available!")
        return self.fail_message is None

    @Slot(str)
    def failure(self, message):
        self.fail_message = message
        self._is_done = True

    @Slot()
    def success(self):
        self._is_done = True


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for DataConnectionExecutable."""
from multiprocessing import Lock
import pathlib
import tempfile
import unittest
from unittest import mock
from spine_engine import ExecutionDirection
from spine_items.data_connection.executable_item import ExecutableItem


class TestDataConnectionExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(ExecutableItem.item_type(), "Data Connection")

    def test_from_dict(self):
        logger = mock.MagicMock()
        item_dict = {
            "type": "Data Connection",
            "description": "",
            "x": 0,
            "y": 0,
            "file_references": [{"type": "path", "relative": True, "path": "/temp/temp1.txt"}],
            "db_references": [],
            "db_credentials": {},
        }
        dc_data_dir = pathlib.Path(self._temp_dir.name, ".spinetoolbox", "items", "dc")
        dc_data_dir.mkdir(parents=True)
        temp_file_path = pathlib.Path(dc_data_dir, "file.txt")
        temp_file_path.touch()
        item = ExecutableItem.from_dict(item_dict, "DC", self._temp_dir.name, None, {}, logger)
        self.assertIsInstance(item, ExecutableItem)
        self.assertEqual("Data Connection", item.item_type())
        self.assertEqual(2, len(item._file_paths))

    def test_stop_execution(self):
        executable = ExecutableItem("name", [], [], self._temp_dir.name, mock.MagicMock())
        with mock.patch(
            "spine_engine.project_item.executable_item_base.ExecutableItemBase.stop_execution"
        ) as mock_stop_execution:
            executable.stop_execution()
            mock_stop_execution.assert_called_once()

    def test_execute(self):
        executable = ExecutableItem("name", [], [], self._temp_dir.name, mock.MagicMock())
        self.assertTrue(executable.execute([], [], Lock()))

    def test_output_resources_backward(self):
        dc_data_dir = pathlib.Path(self._temp_dir.name, ".spinetoolbox", "items", "name")
        dc_data_dir.mkdir(parents=True)
        temp_file_path = pathlib.Path(dc_data_dir, "file.txt")
        temp_file_path.touch()
        executable = ExecutableItem("name", ["file_reference"], [], self._temp_dir.name, mock.MagicMock())
        self.assertEqual(executable.output_resources(ExecutionDirection.BACKWARD), [])

    def test_output_resources_forward(self):
        file_reference = pathlib.Path(self._temp_dir.name, "file_reference")
        file_reference.touch()
        dc_data_dir = pathlib.Path(self._temp_dir.name, ".spinetoolbox", "items", "name")
        dc_data_dir.mkdir(parents=True)
        temp_file_path = pathlib.Path(dc_data_dir, "data_file")
        temp_file_path.touch()
        executable = ExecutableItem("name", [str(file_reference)], [], self._temp_dir.name, mock.MagicMock())
        output_resources = executable.output_resources(ExecutionDirection.FORWARD)
        self.assertEqual(len(output_resources), 2)
        resource = output_resources[0]
        self.assertEqual(resource.type_, "file")
        self.assertTrue(pathlib.Path(resource.path).samefile(file_reference))
        self.assertEqual(resource.metadata, {})
        resource = output_resources[1]
        self.assertEqual(resource.type_, "file")
        self.assertTrue(pathlib.Path(resource.path).samefile(temp_file_path))
        self.assertEqual(resource.metadata, {})


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains common & shared (Q)widgets."""
import os
from PySide6.QtCore import Property, QMimeData, Qt, QUrl, Signal, Slot
from PySide6.QtGui import QDrag, QIntValidator
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QLineEdit,
    QStatusBar,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionComboBox,
    QTreeView,
    QWidget,
)
from spinetoolbox.config import APPLICATION_PATH, STATUSBAR_SS
from spinetoolbox.helpers import get_open_file_name_in_last_dir
from .utils import convert_to_sqlalchemy_url


class ArgsTreeView(QTreeView):
    def dragEnterEvent(self, event):
        super().dragEnterEvent(event)
        event.accept()


class ReferencesTreeView(QTreeView):
    """Custom QTreeView class for Data Connection and View properties."""

    files_dropped = Signal(list)
    del_key_pressed = Signal()

    def __init__(self, parent):
        """Initializes the view.

        Args:
            parent (QWidget): The parent of this view
        """
        super().__init__(parent=parent)

    def dragEnterEvent(self, event):
        """Accepts file drops from the filesystem."""
        urls = event.mimeData().urls()
        for url in urls:
            if not url.isLocalFile():
                event.ignore()
                return
            if not os.path.isfile(url.toLocalFile()):
                event.ignore()
                return
        event.accept()
        event.setDropAction(Qt.LinkAction)

    def dragMoveEvent(self, event):
        """Accepts event."""
        event.accept()

    def dropEvent(self, event):
        """Emits files_dropped signal with a list of files for each dropped url."""
        self.files_dropped.emit([url.toLocalFile() for url in event.mimeData().urls()])

    def keyPressEvent(self, event):
        """Overridden method to make the view support deleting items with a delete key."""
        super().keyPressEvent(event)
        if event.key() == Qt.Key_Delete:
            self.del_key_pressed.emit()


class DataTreeView(QTreeView):
    """Custom QTreeView class for the 'Data' files in DataConnection properties."""

    files_dropped = Signal(list)
    del_key_pressed = Signal()

    def __init__(self, parent):
        """Initializes the view.

        Args:
            parent (QWidget): The parent of this view
        """
        super().__init__(parent=parent)
        self.drag_start_pos = None
        self.drag_indexes = []

    def dragEnterEvent(self, event):
        """Accepts file drops from the filesystem."""
        urls = event.mimeData().urls()
        for url in urls:
            if not url.isLocalFile():
                event.ignore()
                return
            if not os.path.isfile(url.toLocalFile()):
                event.ignore()
                return
        event.accept()
        event.setDropAction(Qt.CopyAction)

    def dragMoveEvent(self, event):
        """Accepts event."""
        event.accept()

    def dropEvent(self, event):
        """Emits files_dropped signal with a list of files for each dropped url."""
        self.files_dropped.emit([url.toLocalFile() for url in event.mimeData().urls()])

    def mousePressEvent(self, event):
        """Registers drag start position."""
        if event.button() == Qt.LeftButton:
            self.drag_start_pos = event.position().toPoint()
            self.drag_indexes = self.selectedIndexes()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """Starts dragging action if needed."""
        if not event.buttons() & Qt.LeftButton:
            return
        if not self.drag_start_pos:
            return
        if not self.drag_indexes:
            return
        if (event.position().toPoint() - self.drag_start_pos).manhattanLength() < QApplication.startDragDistance():
            return
        drag = QDrag(self)
        mimeData = QMimeData()
        urls = []
        for index in self.drag_indexes:
            file_path = index.data(Qt.ItemDataRole.UserRole)
            if not file_path:
                return
            urls.append(QUrl.fromLocalFile(file_path))
        mimeData.setUrls(urls)
        drag.setMimeData(mimeData)
        icon = self.drag_indexes[0].data(Qt.ItemDataRole.DecorationRole)
        if icon:
            pixmap = icon.pixmap(32, 32)
            drag.setPixmap(pixmap)
            drag.setHotSpot(pixmap.rect().center())
        drag.exec()

    def mouseReleaseEvent(self, event):
        """Forgets drag start position"""
        self.drag_start_pos = None
        super().mouseReleaseEvent(event)  # Fixes bug in extended selection

    def keyPressEvent(self, event):
        """Overridden method to make the view support deleting items with a delete key."""
        super().keyPressEvent(event)
        if event.key() == Qt.Key_Delete:
            self.del_key_pressed.emit()


class FilterEditDelegateBase(QStyledItemDelegate):
    """Base class for regexp filter edit delegates.

    Derived classes should reimplement at least ``createEditor()``.
    """

    def updateEditorGeometry(self, editor, option, index):
        top_left = option.rect.topLeft()
        popup_position = editor.parent().mapToGlobal(top_left)
        size_hint = editor.sizeHint()
        editor.setGeometry(
            popup_position.x(), popup_position.y(), max(option.rect.width(), size_hint.width()), size_hint.height()
        )


class FilterEdit(QWidget):
    """Filter regular expression editor."""

    def __init__(self, ui_form, parent):
        """
        Args:
            ui_form (Any): an interface from created from a .ui file
            parent (QWidget):
        """
        super().__init__(parent)
        self.setWindowFlags(Qt.WindowType.Popup)
        self._ui = ui_form
        self._ui.setupUi(self)
        self._focused = False

    def focusInEvent(self, e):
        self._ui.regexp_line_edit.setFocus()
        self._ui.regexp_line_edit.selectAll()
        super().focusInEvent(e)

    def keyPressEvent(self, event):
        # Relay key press events to the regexp line edit. Otherwise we may lose the first letter.
        return self._ui.regexp_line_edit.keyPressEvent(event)

    def regexp(self):
        """Returns the current regular expression.

        Returns:
            str: regular expression
        """
        return self._ui.regexp_line_edit.text()

    def set_regexp(self, regexp):
        """Sets a regular expression for editing.

        Args:
            regexp (str): new regular expression
        """
        self._ui.regexp_line_edit.setText(regexp)

    regexp = Property(str, regexp, set_regexp, user=True)
    """Property used to communicate with the editor delegate."""


class FileDropTargetLineEdit(QLineEdit):
    """A line edit that accepts file drops and displays the path."""

    def dragEnterEvent(self, event):
        """Accept a single file drop from the filesystem."""
        urls = event.mimeData().urls()
        if len(urls) > 1:
            event.ignore()
            return
        url = urls[0]
        if not url.isLocalFile():
            event.ignore()
            return
        if not os.path.isfile(url.toLocalFile()):
            event.ignore()
            return
        event.accept()
        event.setDropAction(Qt.LinkAction)

    def dragMoveEvent(self, event):
        """Accept event."""
        event.accept()

    def dropEvent(self, event):
        """Sets the text to the file path."""
        url = event.mimeData().urls()[0]
        self.setText(url.toLocalFile())


def _set_line_edit_text(edit, text):
    """Sets QLineEdit's text only if it is changing.

    Avoids sudden jumps in cursors when e.g. the latest change goes through the
    undo stack

    Args:
        edit (QLineEdit): line edit
        text (str): new text to set for the edit
    """
    if text != edit.text():
        edit.setText(text)


KNOWN_SQL_DIALECTS = ("mysql", "sqlite", "postgresql")


class UrlSelectorWidget(QWidget):
    """Widget for setting up database URLs."""

    url_changed = Signal()
    """Emitted whenever the URL changes."""

    def __init__(self, parent=None):
        """
        Args:
            parent (QWidget, optional): parent widget
        """
        from .ui.url_selector_widget import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._url = None
        self._get_sqlite_file_path = None
        self._logger = None
        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self._ui.comboBox_dialect.setCurrentIndex(-1)
        self._ui.lineEdit_port.setValidator(QIntValidator())
        self._ui.comboBox_dialect.currentTextChanged.connect(self._enable_dialect)
        self._ui.toolButton_select_sqlite_file.clicked.connect(self._select_sqlite_file)
        self._ui.comboBox_dialect.currentTextChanged.connect(lambda: self.url_changed.emit())
        self._ui.lineEdit_host.textChanged.connect(lambda: self.url_changed.emit())
        self._ui.lineEdit_port.textChanged.connect(lambda: self.url_changed.emit())
        self._ui.lineEdit_database.textChanged.connect(lambda: self.url_changed.emit())
        self._ui.lineEdit_username.textChanged.connect(lambda: self.url_changed.emit())
        self._ui.lineEdit_password.textChanged.connect(lambda: self.url_changed.emit())

    def setup(self, dialects, select_sqlite_file_callback, hide_schema, logger):
        """Sets the widget up for usage.

        Args:
            dialects (Sequence of str): available SQL dialects
            select_sqlite_file_callback (Callable): function that returns a path to SQLite file or None
            hide_schema (bool): True to hide the Schema field
            logger (LoggerInterface): logger
        """
        self._get_sqlite_file_path = select_sqlite_file_callback
        self._logger = logger
        self._ui.comboBox_dialect.addItems(dialects)
        if hide_schema:
            self._ui.schema_line_edit.setVisible(False)
            self._ui.schema_label.setVisible(False)

    def set_url(self, url):
        """Sets the URL for the widget.

        Args:
            url (dict): URL as dict
        """
        dialect = url.get("dialect", "")
        host = url.get("host", "")
        port = url.get("port", "")
        database = url.get("database", "")
        schema = url.get("schema", "")
        username = url.get("username", "")
        password = url.get("password", "")
        self.blockSignals(True)
        if not dialect:
            self._ui.comboBox_dialect.setCurrentIndex(-1 if self._ui.comboBox_dialect.count() == 0 else 0)
        elif dialect != self._ui.comboBox_dialect.currentText():
            self._ui.comboBox_dialect.setCurrentText(dialect)
        _set_line_edit_text(self._ui.lineEdit_host, host)
        _set_line_edit_text(self._ui.lineEdit_port, port)
        _set_line_edit_text(self._ui.lineEdit_database, database)
        _set_line_edit_text(self._ui.schema_line_edit, schema)
        _set_line_edit_text(self._ui.lineEdit_username, username)
        _set_line_edit_text(self._ui.lineEdit_password, password)
        self.blockSignals(False)

    def url_dict(self):
        """Returns the URL as dictionary.

        Returns:
            dict: URL as dict
        """
        return {
            "dialect": self._ui.comboBox_dialect.currentText(),
            "host": self._ui.lineEdit_host.text(),
            "port": self._ui.lineEdit_port.text(),
            "database": self._ui.lineEdit_database.text(),
            "schema": self._ui.schema_line_edit.text(),
            "username": self._ui.lineEdit_username.text(),
            "password": self._ui.lineEdit_password.text(),
        }

    @Slot(bool)
    def _select_sqlite_file(self, _=False):
        """Select SQLite file."""
        file_path = self._get_sqlite_file_path()
        if file_path is not None:
            self._ui.lineEdit_database.setText(file_path)

    @Slot(str)
    def _enable_dialect(self, dialect):
        """Enables the given dialect in the item controls.

        Args:
            dialect (str): SQL dialect
        """
        if dialect == "":
            self.enable_no_dialect()
        elif dialect == "sqlite":
            self.enable_sqlite()
        else:
            self.enable_common()

    def enable_no_dialect(self):
        """Adjusts widget enabled status to default when no dialect is selected."""
        self._ui.comboBox_dialect.setEnabled(True)
        self._ui.toolButton_select_sqlite_file.setEnabled(False)
        self._ui.lineEdit_host.setEnabled(False)
        self._ui.lineEdit_port.setEnabled(False)
        self._ui.lineEdit_database.setEnabled(False)
        self._ui.lineEdit_username.setEnabled(False)
        self._ui.lineEdit_password.setEnabled(False)
        self._ui.schema_line_edit.setEnabled(False)

    def enable_sqlite(self):
        """Adjusts controls to sqlite connection specification."""
        self._ui.toolButton_select_sqlite_file.setEnabled(True)
        self._ui.lineEdit_host.setEnabled(False)
        self._ui.lineEdit_port.setEnabled(False)
        self._ui.lineEdit_database.setEnabled(True)
        self._ui.lineEdit_username.setEnabled(False)
        self._ui.lineEdit_password.setEnabled(False)
        self._ui.schema_line_edit.setEnabled(False)
        self._ui.lineEdit_host.clear()
        self._ui.lineEdit_port.clear()
        self._ui.lineEdit_username.clear()
        self._ui.lineEdit_password.clear()

    def enable_common(self):
        """Adjusts controls to 'common' connection specification."""
        self._ui.toolButton_select_sqlite_file.setEnabled(False)
        self._ui.lineEdit_host.setEnabled(True)
        self._ui.lineEdit_port.setEnabled(True)
        self._ui.lineEdit_database.setEnabled(True)
        self._ui.lineEdit_username.setEnabled(True)
        self._ui.lineEdit_password.setEnabled(True)
        self._ui.schema_line_edit.setEnabled(True)


class UrlSelectorDialog(QDialog):
    msg_error = Signal(str)

    def __init__(self, app_settings, hide_schema, logger, parent=None):
        """
        Args:
            app_settings (QSettings): Toolbox settings
            hide_schema (bool): if True, hide the Schema field
            logger (LoggerInterface): logger
            parent (QWidget, optional): parent widget
        """
        from .ui.url_selector_dialog import Ui_Dialog  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._sa_url = None
        self._app_settings = app_settings
        self._logger = logger
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.ui.url_selector_widget.setup(KNOWN_SQL_DIALECTS, self._browse_sqlite_file, hide_schema, self._logger)
        self.ui.url_selector_widget.url_changed.connect(self._refresh_url)
        # Add status bar to form
        self.statusbar = QStatusBar(self)
        self.statusbar.setFixedHeight(20)
        self.statusbar.setSizeGripEnabled(False)
        self.statusbar.setStyleSheet(STATUSBAR_SS)
        self.ui.horizontalLayout_statusbar_ph.addWidget(self.statusbar)
        self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)
        self.msg_error.connect(self.statusbar.showMessage)

    @property
    def url(self):
        if self._sa_url is None:
            return ""
        return str(self._sa_url)

    def url_dict(self):
        return self.ui.url_selector_widget.url_dict()

    def set_url_dict(self, url):
        """Sets the URL.

        Args:
            url (dict): URL as dict
        """
        self.ui.url_selector_widget.set_url(url)

    @property
    def dialect(self):
        return self.ui.url_selector_widget.url_dict().get("dialect")

    @Slot()
    def _refresh_url(self):
        """Updates the URL widget and status bar."""
        url = self.ui.url_selector_widget.url_dict()
        self._sa_url = convert_to_sqlalchemy_url(url, logger=self)
        if self._sa_url is not None:
            self.statusbar.clearMessage()
        self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(self._sa_url is not None)

    def _browse_sqlite_file(self):
        """Opens a browser to select a SQLite file.

        Returns:
            str: path to the file or None if operation was cancelled
        """
        filter_ = "*.sqlite;;*.*"
        key = "selectImportSourceSQLiteFile"
        filepath, _ = get_open_file_name_in_last_dir(
            self._app_settings, key, self, "Select an SQLite file", APPLICATION_PATH, filter_=filter_
        )
        return filepath if filepath else None


def combo_box_width(font_metric_widget, items):
    """Returns section width.

    Args:
        font_metric_widget (QWidget): Widget whose font metrics are used
        items (Iterable of str): combo box items

    Returns:
        int: width of a combo box containing the given items
    """
    fm = font_metric_widget.fontMetrics()
    style = QApplication.instance().style()
    option = QStyleOptionComboBox()
    rect = style.subControlRect(QStyle.ComplexControl.CC_ComboBox, option, QStyle.SubControl.SC_ComboBoxArrow)
    return max(fm.horizontalAdvance(item) for item in items) + rect.width()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Module for tool icon class."""
from PySide6.QtCore import QPointF, QTimeLine, Slot
from spinetoolbox.project_item_icon import ProjectItemIcon
from ..animations import AnimationSignaller


class ToolIcon(ProjectItemIcon):
    def __init__(self, toolbox, icon, icon_color):
        """Tool icon for the Design View.

        Args:
            toolbox (ToolBoxUI): QMainWindow instance
            icon (str): icon resource path
            icon_color (QColor): Icon's color
        """
        super().__init__(toolbox, icon, icon_color)
        self.time_line = QTimeLine()
        self.time_line.setLoopCount(0)  # loop forever
        self.time_line.setFrameRange(0, 10)
        self.time_line.setDuration(1200)
        self.time_line.setDirection(QTimeLine.Direction.Backward)
        self.time_line.valueChanged.connect(self._handle_time_line_value_changed)
        self.time_line.stateChanged.connect(self._handle_time_line_state_changed)
        self._svg_item_pos = self.svg_item.pos()
        rect = self.svg_item.sceneBoundingRect()
        self._anim_transformation_origin_point_y = -0.75 * rect.height()
        self._anim_delta_x_factor = 0.5 * rect.width()
        self.animation_signaller = AnimationSignaller()
        self.animation_signaller.animation_started.connect(self.start_animation)
        self.animation_signaller.animation_stopped.connect(self.stop_animation)

    @Slot(float)
    def _handle_time_line_value_changed(self, value):
        angle = value * 45.0
        self.svg_item.setRotation(angle)
        delta_y = 0.5 * self.svg_item.sceneBoundingRect().height()
        delta = QPointF(self._anim_delta_x_factor * value, delta_y)
        self.svg_item.setPos(self._svg_item_pos + delta)

    @Slot(QTimeLine.State)
    def _handle_time_line_state_changed(self, new_state):
        if new_state == QTimeLine.State.Running:
            self.svg_item.setTransformOriginPoint(0, self._anim_transformation_origin_point_y)
        elif new_state == QTimeLine.State.NotRunning:
            self.svg_item.setTransformOriginPoint(0, 0)
            self.svg_item.setPos(self._svg_item_pos)
            self.svg_item.setRotation(0)

    @Slot()
    def start_animation(self):
        """Starts the item execution animation."""
        if self.time_line.state() == QTimeLine.State.Running:
            return
        self.time_line.start()

    @Slot()
    def stop_animation(self):
        """Stop animation"""
        if self.time_line.state() != QTimeLine.State.Running:
            return
        self.time_line.stop()

    def mouseDoubleClickEvent(self, e):
        """Opens Tool Specification editor when this Tool icon is double-clicked.

        Args:
            e (QGraphicsSceneMouseEvent): Event
        """
        super().mouseDoubleClickEvent(e)
        item = self._toolbox.project().get_item(self._name)
        item.show_specification_window()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for the ``import_mapping_options`` module."""
import unittest
from PySide6.QtCore import QItemSelectionModel, Qt
from PySide6.QtGui import QUndoStack
from PySide6.QtWidgets import QApplication
from spine_items.importer.mvcmodels.mappings_model import MappingsModel
from spine_items.importer.mvcmodels.mappings_model_roles import Role
from spine_items.importer.ui.import_editor_window import Ui_MainWindow
from spine_items.importer.widgets.import_mapping_options import ImportMappingOptions
from spinetoolbox.helpers import signal_waiter
from tests.mock_helpers import parent_widget


class TestImportMappingOptions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_time_series_repeat_flag_check_box(self):
        with parent_widget() as parent:
            undo_stack = QUndoStack()
            mappings_model = MappingsModel(undo_stack, parent)
            mappings_dict = self._template_mapping()
            mappings_model.restore(mappings_dict)
            ui = Ui_MainWindow()
            ui.setupUi(parent)
            ui.source_list.setModel(mappings_model)
            ui.mapping_list.setModel(mappings_model)
            ui.mapping_list.setRootIndex(mappings_model.dummy_parent())
            _ = ImportMappingOptions(mappings_model, ui, undo_stack)
            table_index = mappings_model.index(1, 0)
            ui.source_list.selectionModel().setCurrentIndex(table_index, QItemSelectionModel.ClearAndSelect)
            mapping_list_index = mappings_model.index(0, 0, table_index)
            ui.mapping_list.selectionModel().setCurrentIndex(mapping_list_index, QItemSelectionModel.ClearAndSelect)
            self.assertEqual(ui.time_series_repeat_check_box.checkState(), Qt.CheckState.Unchecked)
            table_index = mappings_model.index(1, 0)
            flattened_mappings = mappings_model.index(0, 0, table_index).data(Role.FLATTENED_MAPPINGS)
            self.assertEqual(flattened_mappings.value_mapping().options.get("repeat", False), False)
            with signal_waiter(ui.time_series_repeat_check_box.stateChanged) as waiter:
                ui.time_series_repeat_check_box.setChecked(True)
                waiter.wait()
            self.assertEqual(ui.time_series_repeat_check_box.checkState(), Qt.CheckState.Checked)
            table_index = mappings_model.index(1, 0)
            flattened_mappings = mappings_model.index(0, 0, table_index).data(Role.FLATTENED_MAPPINGS)
            self.assertEqual(flattened_mappings.value_mapping().options["repeat"], True)

    @staticmethod
    def _template_mapping():
        return {
            "table_mappings": {
                "Sheet1": [
                    {
                        "Mapping 1": {
                            "mapping": [
                                {"map_type": "ObjectClass", "position": "hidden", "value": "Item"},
                                {"map_type": "Object", "position": "hidden", "value": "spoon"},
                                {"map_type": "ObjectMetadata", "position": "hidden"},
                                {"map_type": "ParameterDefinition", "position": "hidden", "value": "temperature"},
                                {"map_type": "Alternative", "position": "hidden", "value": "Base"},
                                {"map_type": "ParameterValueMetadata", "position": "hidden"},
                                {"map_type": "ParameterValueType", "position": "hidden", "value": "time_series"},
                                {"map_type": "IndexName", "position": "hidden"},
                                {"map_type": "ParameterValueIndex", "position": 0},
                                {"map_type": "ExpandedValue", "position": 1},
                            ]
                        }
                    }
                ]
            },
            "selected_tables": ["Sheet1"],
            "table_options": {"Sheet1": {}},
            "table_types": {"Sheet1": {"0": "datetime", "1": "float"}},
            "table_default_column_type": {},
            "table_row_types": {},
            "source_type": "ExcelConnector",
        }


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for :class:`IntegerSequenceDateTimeConvertSpecDialog`."""
import unittest
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication
from spine_items.importer.widgets.table_view_with_button_header import IntegerSequenceDateTimeConvertSpecDialog
from spinedb_api import DateTime, Duration
from spinedb_api.import_mapping.type_conversion import IntegerSequenceDateTimeConvertSpec


class TestIntegerSequenceDateTimeConvertSpecDialog(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_restore_previous_spec(self):
        spec = IntegerSequenceDateTimeConvertSpec(DateTime("2020-11-12T15:45"), 23, Duration("5h"))
        widget = IntegerSequenceDateTimeConvertSpecDialog(spec, None)
        self.assertEqual(widget.datetime.dateTime().toString(Qt.ISODate), "2020-11-12T15:45:00")
        self.assertEqual(widget.start_integer.value(), 23)
        self.assertEqual(widget.duration.text(), "5h")


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains a model to handle source tables and import mapping."""
from enum import Enum, unique
from spinedb_api import InvalidMapping, ParameterValueFormatError, from_database
from spinedb_api.helpers import fix_name_ambiguity
from spinedb_api.import_mapping.import_mapping import (
    AlternativeMapping,
    DefaultValueIndexNameMapping,
    DimensionMapping,
    ElementMapping,
    EntityAlternativeActivityMapping,
    EntityClassMapping,
    EntityGroupMapping,
    EntityMapping,
    EntityMetadataMapping,
    IndexNameMapping,
    ParameterDefaultValueIndexMapping,
    ParameterDefaultValueMapping,
    ParameterDefaultValueTypeMapping,
    ParameterDefinitionMapping,
    ParameterValueIndexMapping,
    ParameterValueListMapping,
    ParameterValueMapping,
    ParameterValueTypeMapping,
    ScenarioAlternativeMapping,
    ScenarioMapping,
    check_validity,
)
from spinedb_api.mapping import (
    Position,
    is_pivoted,
    parse_fixed_position_value,
    unflatten,
    unparse_fixed_position_value,
)
from spinedb_api.parameter_value import split_value_and_type
from spinetoolbox.helpers import color_from_index
from spinetoolbox.spine_db_manager import SpineDBManager


@unique
class MappingType(Enum):
    EntityClass = "Entity class"
    EntityGroup = "Entity group"
    Alternative = "Alternative"
    Scenario = "Scenario"
    ScenarioAlternative = "Scenario alternative"
    ParameterValueList = "Parameter value list"


VALUE_TYPES = {
    "Single value": "single_value",
    "Array": "array",
    "Map": "map",
    "Time series": "time_series",
    "Time pattern": "time_pattern",
}
DISPLAY_VALUE_TYPES = {v: k for k, v in VALUE_TYPES.items()}

DISPLAY_MAPPING_NAMES = {
    "EntityClass": "Entity class names",
    "Entity": "Entity names",
    "EntityMetadata": "Entity metadata",
    "EntityGroup": "Member names",
    "Dimension": "Dimension names",
    "Element": "Element names",
    "EntityAlternativeActivity": "Entity activities",
    "Alternative": "Alternative names",
    "Scenario": "Scenario names",
    "ScenarioActiveFlag": "Scenario active flags",
    "ScenarioAlternative": "Alternative names",
    "ScenarioBeforeAlternative": "Before alternative names",
    "ParameterValueList": "Value list names",
    "ParameterValueListValue": "Parameter values",
    "ParameterDefinition": "Parameter names",
    "ParameterValue": "Parameter values",
    "ParameterValueType": None,
    "ParameterValueMetadata": "Parameter value metadata",
    "IndexName": "Parameter index names",
    "ParameterValueIndex": "Parameter indexes",
    "ExpandedValue": "Parameter values",
    "ParameterDefaultValue": "Parameter default values",
    "ParameterDefaultValueType": None,
    "DefaultValueIndexName": "Parameter index names",
    "ParameterDefaultValueIndex": "Parameter indexes",
    "ExpandedDefaultValue": "Parameter default values",
}


class FlattenedMappings:
    def __init__(self, root_mapping):
        """
        Args:
            root_mapping (ImportMapping): root mapping
        """
        self._components = None
        self._map_type = None
        self._value_type = None
        self._display_names = None
        self._display_to_logical = None
        self._display_colors = None
        self._row_issues = None
        self.mapping_list_item = None
        self.set_root_mapping(root_mapping)

    @property
    def map_type(self):
        return self._map_type

    def set_root_mapping(self, root_mapping):
        """Sets root mapping resetting data.

        Args:
            root_mapping (ImportMapping)
        """
        self._components = root_mapping.flatten()
        self._map_type = self._resolve_map_type(self._components)
        self._value_type = self._resolve_value_type()
        self._display_names, self._display_to_logical = self._displayable_component_names()
        self._display_colors = self._make_colors()
        self._row_issues = None

    @property
    def display_names(self):
        return self._display_names

    @property
    def display_colors(self):
        return self._display_colors

    @property
    def root_mapping(self):
        return self._components[0]

    @property
    def value_type(self):
        return self._value_type

    def is_time_series_value(self):
        """Checks if value is a time series.

        Returns:
            bool: True if value is a time series, False otherwise
        """
        return self._value_type == "Time series"

    def is_map_value(self):
        """Checks if value is a map.

        Returns:
            bool: True if value is a map, False otherwise
        """
        return self._value_type == "Map"

    def map_dimension_count(self):
        """Calculates dimensions count for Map type value.

        Returns:
            int: dimension count
        """
        if not self.is_map_value():
            return 1
        parameter_type = self.display_parameter_type()
        index_mapping = ParameterValueIndexMapping if parameter_type == "Value" else ParameterDefaultValueIndexMapping
        return len([m for m in self._components if isinstance(m, index_mapping)])

    def set_map_dimension_count(self, dimension_count):
        """Sets new dimensions for Map type value.

        Args:
            dimension_count (int): dimension count
        """
        if not self.is_map_value():
            return
        previous_dimension_count = self.map_dimension_count()
        parameter_type = self.display_parameter_type()
        name_mapping_cls = IndexNameMapping if parameter_type == "Value" else DefaultValueIndexNameMapping
        index_mapping_cls = (
            ParameterValueIndexMapping if parameter_type == "Value" else ParameterDefaultValueIndexMapping
        )
        last_index_mapping = next(m for m in reversed(self._components) if isinstance(m, index_mapping_cls))
        last_index_mapping_child = last_index_mapping.child
        if dimension_count > previous_dimension_count:
            for _ in range(previous_dimension_count, dimension_count):
                name_mapping = name_mapping_cls(Position.hidden)
                last_index_mapping.child = name_mapping
                index_mapping = index_mapping_cls(Position.hidden)
                name_mapping.child = index_mapping
                last_index_mapping = index_mapping
        else:
            for _ in range(dimension_count, previous_dimension_count):
                name_mapping = last_index_mapping.parent
                last_index_mapping = name_mapping.parent
        last_index_mapping.child = last_index_mapping_child
        self.set_root_mapping(self._components[0])

    def set_map_compress(self, compress):
        """
        Sets the compress flag for Map type parameters.

        Args:
            compress (bool): flag value
        """
        if not self.is_map_value():
            return
        self.value_mapping().compress = compress

    def set_value_components(self, component):
        """Sets value mapping components.

        Args:
            component (ImportMapping): parameter value component root
        """
        m = self.value_mapping()
        if m is None or m.parent is None:
            return
        m.parent.child = component
        self.set_root_mapping(self._components[0])

    def append_tail_component(self, component):
        """Adds a new component as the last component in mappings.

        Args:
            component (ImportMapping): component to add
        """
        self._components[-1].child = component
        self.set_root_mapping(self._components[0])

    def cut_tail_component(self):
        """Removes the tail component."""
        self._components[-2].child = None
        self.set_root_mapping(self._components[0])

    def read_start_row(self):
        """Returns read start row.

        Returns:
            int: read start row
        """
        return self._components[0].read_start_row

    def set_read_start_row(self, row):
        """Sets read start row.

        Args:
            row (int): read start row
        """
        self._components[0].read_start_row = row

    def skip_columns(self):
        """Returns skipped columns.

        Returns:
            list: list of skipped column indexes or names
        """
        return self._components[0].skip_columns

    def set_skip_columns(self, skip_columns):
        """Sets skipped columns.

        Args:
            skip_columns (list): skipped column indexes or names
        """
        self._components[0].skip_columns = skip_columns

    def component_at(self, row):
        """Returns mapping corresponding to a display row.

        Args:
            row (int): display row index

        Returns:
            ImportMapping: mapping
        """
        return self._components[self._display_to_logical[row]]

    def display_position_type(self, row):
        """Creates a 'type' for position for displaying.

        Args:
            row (int): display row

        Returns:
            str: display position
        """
        if self.is_component_pivoted(row):
            return "Pivoted"
        component = self._components[self._display_to_logical[row]]
        if component.position == Position.hidden:
            if component.value is None:
                return "None"
            return "Constant"
        if component.position == Position.header:
            if component.value is None:
                return "Headers"
            return "Column Header"
        if component.position == Position.table_name:
            return "Table Name"
        if component.position == Position.mapping_name:
            return "Mapping Name"
        if component.position == Position.fixed:
            return "Fixed"
        if component.position >= 0:
            return "Column"
        return "Row"

    def is_component_pivoted(self, row):
        return (
            len(self._components) > 1
            and any(
                is_pivoted(c.position) or (c.position == Position.header and c.value is None)
                for c in self._components[:-1]
            )
            and row == len(self._display_names) - 1
        )

    def set_display_position_type(self, row, position_type):
        """Sets component's position 'type'.

        Args:
            row (int): display row
            position_type (str or int): position type
        """
        component = self.component_at(row)
        if position_type in ("None", "", None):
            component.position = Position.hidden
            component.value = None
        elif position_type == "Constant":
            component.position = Position.hidden
            component.value = "constant"
        elif position_type == "Column":
            component.position = 0
            component.value = None
        elif position_type == "Column Header":
            component.position = Position.header
            component.value = 0
        elif position_type == "Headers":
            component.position = Position.header
            component.value = None
        elif position_type == "Row":
            component.position = -1
            component.value = None
        elif position_type == "Table Name":
            component.position = Position.table_name
        elif position_type == "Mapping Name":
            component.position = Position.mapping_name
        elif position_type == "Fixed":
            component.position = Position.fixed
            component.value = "table_name: row, column"
        self._row_issues = None

    def display_position(self, row):
        """Converts a position to something displayable.

        Args:
            row (int): display row index

        Returns:
            str or int or bool: display position
        """
        component = self._components[self._display_to_logical[row]]
        # A) Handle two special cases for value mappings
        if self.is_component_pivoted(row):
            # 1. Pivoted data
            return "Pivoted values"
        if self._display_names[row].endswith("values"):
            if component.position == Position.hidden and component.value is not None:
                # 2. Constant value: we want special database value support
                try:
                    value = from_database(*split_value_and_type(component.value))
                    return SpineDBManager.display_data_from_parsed(value)
                except ParameterValueFormatError:
                    return None
        # B) Handle all other cases
        if component.position == Position.hidden:
            return component.value
        if component.position == Position.header:
            if component.value is None:
                return "Headers"
            return component.value + 1
        if component.position == Position.table_name:
            return "<table name>"
        if component.position == Position.mapping_name:
            return "<mapping name>"
        if component.position == Position.fixed:
            try:
                table_name, row, column = parse_fixed_position_value(component.value)
            except InvalidMapping:
                return component.value
            return unparse_fixed_position_value(table_name, row + 1, column + 1)
        if component.position >= 0:
            return component.position + 1
        return -(component.position + 1) + 1

    def set_display_position(self, row, position_type, position):
        """Sets component's position.

        Args:
            row (int): display row
            position_type (str): display position type
            position (str or int): display position
        """
        component = self.component_at(row)
        if position_type == "Constant":
            component.value = position
        elif position_type == "Column":
            component.position = position - 1
        elif position_type == "Column Header":
            component.value = position - 1
        elif position_type == "Row":
            component.position = -position
        if position_type == "Fixed":
            try:
                table_name, row, column = parse_fixed_position_value(position)
            except InvalidMapping:
                component.value = position
            else:
                component.value = unparse_fixed_position_value(table_name, row - 1, column - 1)
        self._row_issues = None

    def display_row_issues(self, row):
        """Validates a display row for issues.

        Args:
            row (int): display row index

        Returns:
            list of str: issues
        """
        if self._row_issues is None:
            self._row_issues = {}
            for issue in check_validity(self._components[0]):
                self._row_issues.setdefault(issue.rank, []).append(issue.msg)
        return self._row_issues.get(self._display_to_logical[row], [])

    def _component_name_from_type(self, component_type):
        """Computes display name for a mapping component.

        Args:
            component_type (str): component's MAP_TYPE

        Returns:
            str: display name or None if component is invisible
        """
        if self._map_type == MappingType.EntityGroup and component_type == "Entity":
            return "Group names"
        if self._value_type == "Array" and component_type in ("ParameterValueIndex", "ParameterDefaultValueIndex"):
            return None
        if self._map_type == MappingType.EntityClass and component_type == "Entity" and self.has_dimensions():
            return None
        return DISPLAY_MAPPING_NAMES[component_type]

    def _displayable_component_names(self):
        """Generates a list of displayable mapping component names.

        Returns:
            list of str: displayable names
        """
        component_names = [self._component_name_from_type(m.MAP_TYPE) for m in self._components]
        display_names = []
        display_row = 0
        display_row_to_flattened_index = {}
        for logical_row, name in enumerate(component_names):
            if name is None:
                continue
            display_names.append(name)
            display_row_to_flattened_index[display_row] = logical_row
            display_row += 1
        display_names = fix_name_ambiguity(display_names, prefix=" ")
        return display_names, display_row_to_flattened_index

    @staticmethod
    def _resolve_map_type(flattened):
        """Computes general map type for given flattened mappings.

        Args:
            flattened (list of ImportMapping): flattened mappings

        Returns:
            MappingType: map type
        """
        if not flattened:
            return None
        head_mapping = flattened[0]
        if isinstance(head_mapping, EntityClassMapping):
            if any(isinstance(m, EntityGroupMapping) for m in flattened[1:]):
                return MappingType.EntityGroup
            return MappingType.EntityClass
        if isinstance(head_mapping, AlternativeMapping):
            return MappingType.Alternative
        if isinstance(head_mapping, ScenarioMapping):
            if any(isinstance(m, ScenarioAlternativeMapping) for m in flattened[1:]):
                return MappingType.ScenarioAlternative
            return MappingType.Scenario
        if isinstance(head_mapping, ParameterValueListMapping):
            return MappingType.ParameterValueList

    def _resolve_value_type(self):
        """Computes display name for value type.

        Returns:
            str: value type's display name
        """
        m = self.value_mapping()
        if m is None:
            return None
        if isinstance(m, (ParameterDefaultValueTypeMapping, ParameterValueTypeMapping)):
            return DISPLAY_VALUE_TYPES.get(m.value, "Single value")
        return "Single value"

    def can_have_dimensions(self):
        """Returns True if the mappings can have dimensions.

        Returns:
            bool: True if mappings have dimensions, False otherwise
        """
        return self._map_type == MappingType.EntityClass

    def has_dimensions(self):
        """Returns True if the mappings have dimensions.

        Returns:
            bool: True if mappings have dimensions, False otherwise
        """
        return self.dimension_count() > 0

    def dimension_count(self):
        """Counts entity dimensions.

        Returns:
            int: entity dimensions
        """
        return len([m for m in self._components if isinstance(m, DimensionMapping)])

    def set_dimension_count(self, dimension_count):
        """Sets the numbers of entity dimensions.

        Args:
            dimension_count (int): new dimension count
        """
        if not self.can_have_dimensions():
            return
        current_dimension_count = self.dimension_count()
        last_dim_mapping = next(
            m for m in reversed(self._components) if isinstance(m, (DimensionMapping, EntityClassMapping))
        )
        last_el_mapping = next(m for m in reversed(self._components) if isinstance(m, (ElementMapping, EntityMapping)))
        last_dim_mapping_child = last_dim_mapping.child
        last_el_mapping_child = last_el_mapping.child
        if dimension_count > current_dimension_count:
            for _ in range(current_dimension_count, dimension_count):
                last_dim_mapping.child = DimensionMapping(Position.hidden)
                last_dim_mapping = last_dim_mapping.child
                last_el_mapping.child = ElementMapping(Position.hidden)
                last_el_mapping = last_el_mapping.child
        else:
            for _ in range(dimension_count, current_dimension_count):
                last_dim_mapping = last_dim_mapping.parent
                last_el_mapping = last_el_mapping.parent
        last_dim_mapping.child = last_dim_mapping_child
        last_el_mapping.child = last_el_mapping_child
        self.set_root_mapping(self._components[0])
        self._ensure_consistent_import_entities()

    def may_import_entity_alternatives(self):
        """Checks if the mappings can optionally import entity alternatives.

        Returns:
            bool: True if mappings can import entity alternatives, False otherwise
        """
        return self._map_type == MappingType.EntityClass

    def import_entity_alternatives(self):
        """Returns the import entity alternatives flag.

        Returns:
            bool: True if import entity alternatives is set, False otherwise
        """
        return any(isinstance(m, EntityAlternativeActivityMapping) for m in self._components)

    def set_import_entity_alternatives(self, import_entity_alternatives):
        """Adds or removes AlternativeMapping and EntityAlternativeActivityMapping.

        Legacy mappings may already have AlternativeMapping as part of parameter value mappings;
        in this case we move the AlternativeMapping before EntityAlternativeActivityMapping.

        Args:
            import_entity_alternatives (bool): if True, mappings will be added; if False they will be removed
        """
        if not self.may_import_entity_alternatives():
            return
        if import_entity_alternatives:
            for i in range(len(self._components)):  # pylint: disable=consider-using-enumerate
                if isinstance(self._components[i], AlternativeMapping):
                    alternative_mapping = self._components.pop(i)
                    break
            else:
                alternative_mapping = AlternativeMapping(Position.hidden)
            for i, m in enumerate(self._components):
                if isinstance(m, EntityMetadataMapping):
                    insertion_point = i
                    break
            else:
                raise RuntimeError("Logic error: expected to find EntityMetadataMapping")
            new_mappings = []
            for i, m in enumerate(self._components):
                new_mappings.append(m)
                if i == insertion_point:
                    new_mappings.append(alternative_mapping)
                    new_mappings.append(EntityAlternativeActivityMapping(Position.hidden))
        else:
            new_mappings = []
            has_value = any(isinstance(m, (ParameterValueMapping, ParameterValueTypeMapping)) for m in self._components)
            for m in self._components:
                if isinstance(m, EntityAlternativeActivityMapping):
                    continue
                if isinstance(m, AlternativeMapping) and not has_value:
                    continue
                new_mappings.append(m)
        self.set_root_mapping(unflatten(new_mappings))

    def may_import_entities(self):
        """Checks if the mappings can optionally import entities.

        Returns:
            bool: True if mappings can import entities, False otherwise
        """
        if self._map_type == MappingType.EntityGroup:
            return True
        if self._map_type == MappingType.EntityClass:
            return bool(self._import_entities_mappings())
        return False

    def _import_entities_mappings(self):
        """Collects a list of mapping components that have an import_entities attribute.

        Returns:
            list of ImportMapping: list of components with import_entities
        """
        return [m for m in self._components if hasattr(m, "import_entities")]

    def import_entities(self):
        """Returns the import entities flag.

        Returns:
            bool: True if imports entities is set, False otherwise
        """
        import_enabled_mappings = self._import_entities_mappings()
        return bool(import_enabled_mappings) and all(m.import_entities for m in self._import_entities_mappings())

    def set_import_entities(self, import_entities):
        """Sets the import entities flag for components that support it.

        Args:
            import_entities (bool): flag value
        """
        for m in self._import_entities_mappings():
            m.import_entities = import_entities

    def _ensure_consistent_import_entities(self):
        """If any mapping has the import entities flag set, sets the flag also for all other mappings."""
        mappings = self._import_entities_mappings()
        if mappings and any(mapping.import_entities for mapping in mappings):
            for m in mappings:
                m.import_entities = True

    def has_parameters(self):
        """Returns True if the mappings have parameters.

        Returns:
            bool: True if mapping has parameter, False otherwise
        """
        return self._map_type == MappingType.EntityClass

    def _parameter_definition_component(self):
        """Searches for ParameterDefinitionMapping within the components.

        Returns:
            ParameterDefinitionMapping: mapping component or None if not found
        """
        return next((m for m in self._components if isinstance(m, ParameterDefinitionMapping)), None)

    def set_parameter_components(self, parameter_definition_component):
        """Changes parameter type.

        Args:
            parameter_definition_component (ImportMapping, optional): root of parameter mappings;
                None removes the mappings
        """
        if parameter_definition_component is not None and any(
            isinstance(m, AlternativeMapping) for m in self._components
        ):
            parameter_definition_component = unflatten(
                filter(lambda m: not isinstance(m, AlternativeMapping), parameter_definition_component.flatten())
            )
        m = self._parameter_definition_component()
        parent = m.parent if m is not None else self._components[-1]
        parent.child = parameter_definition_component
        if parameter_definition_component is None and not self.import_entity_alternatives():
            for c in self._components:
                if isinstance(c, AlternativeMapping):
                    c.parent.child = c.child
                    break
        self.set_root_mapping(self._components[0])

    def display_parameter_type(self):
        """Returns a string representation of mappings' parameter type.

        Returns:
            str: parameter type
        """
        if any(isinstance(m, (ParameterValueMapping, ParameterValueTypeMapping)) for m in self._components):
            return "Value"
        if any(isinstance(m, ParameterDefinitionMapping) for m in self._components):
            return "Definition"
        return "None"

    def has_value_component(self):
        """Returns True if any of the components is a value mapping.

        Returns:
            bool: True if components contain a value mapping, False otherwise
        """
        return self.value_mapping() is not None

    def value_type_label(self):
        """Returns a type label for value.

        Returns:
            str: type label
        """
        value_mapping = self.value_mapping()
        if value_mapping is None:
            return "<no label>"
        if isinstance(value_mapping, (ParameterValueTypeMapping, ParameterValueMapping)):
            return "Value:"
        return "Default value:"

    def value_mapping(self):
        """Finds first parameter value mapping component from flattened mappings.

        Returns:
            ImportMapping: value mapping or None if not found
        """
        value_mappings = (
            ParameterValueMapping,
            ParameterValueTypeMapping,
            ParameterDefaultValueMapping,
            ParameterDefaultValueTypeMapping,
        )
        return next((m for m in self._components if isinstance(m, value_mappings)), None)

    def _make_colors(self):
        """Creates display colors for mapping components.

        Returns:
            list of QColor: colors
        """
        component_count = len(self._display_names)
        return [color_from_index(i, component_count).lighter() for i in range(component_count)]



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Classes for handling models in PySide6's model/view framework."""
from collections import namedtuple
from collections.abc import Iterable
from PySide6.QtCore import QModelIndex, QPoint, Qt, Signal, Slot
from PySide6.QtGui import QAction, QCursor, QFont, QIcon
from PySide6.QtWidgets import (  # pylint: disable=unused-import
    QDateTimeEdit,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QHeaderView,
    QLineEdit,
    QMenu,
    QSpinBox,
    QTableView,
    QToolButton,
    QVBoxLayout,
    QWidget,
)
from spinedb_api import DateTime, Duration, ParameterValueFormatError
from spinedb_api.import_mapping.type_conversion import (
    BooleanConvertSpec,
    DateTimeConvertSpec,
    DurationConvertSpec,
    FloatConvertSpec,
    IntegerSequenceDateTimeConvertSpec,
    StringConvertSpec,
    value_to_convert_spec,
)
from spinedb_api.spine_io.importers.reader import TYPE_STRING_TO_CLASS
from spinetoolbox.font import TOOLBOX_FONT
from spinetoolbox.helpers import CharIconEngine
from ..commands import SetColumnOrRowType
from ..mvcmodels.source_data_table_model import SourceDataTableModel

_ALLOWED_TYPES = list(sorted(TYPE_STRING_TO_CLASS.keys()))
_ALLOWED_TYPES.append("integer sequence datetime")

TYPE_TO_FONT_AWESOME_ICON = {
    IntegerSequenceDateTimeConvertSpec.DISPLAY_NAME: chr(int("f073", 16)),
    BooleanConvertSpec.DISPLAY_NAME: chr(int("f6ad", 16)),
    StringConvertSpec.DISPLAY_NAME: chr(int("f031", 16)),
    DateTimeConvertSpec.DISPLAY_NAME: chr(int("f073", 16)),
    DurationConvertSpec.DISPLAY_NAME: chr(int("f017", 16)),
    FloatConvertSpec.DISPLAY_NAME: chr(int("f534", 16)),
}

Margin = namedtuple("Margin", ("left", "right", "top", "bottom"))


class IntegerSequenceDateTimeConvertSpecDialog(QDialog):
    def __init__(self, convert_spec, parent):
        """
        Args:
            convert_spec (IntegerSequenceDateTimeConvertSpec, optional): an existing conversion specification
            parent (QWidget, optional): parent widget
        """
        super().__init__(parent)
        self.setWindowFlag(Qt.WindowType.WindowContextHelpButtonHint, False)
        self.setWindowTitle("New integer sequence datetime")

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.datetime = QDateTimeEdit()
        self.start_integer = QSpinBox()
        self.start_integer.setMaximum(2147483647)
        self.duration = QLineEdit("1h")
        if convert_spec is not None:
            self.datetime.setDateTime(convert_spec.start_datetime.value)
            self.start_integer.setValue(convert_spec.start_int)
            self.duration.setText(str(convert_spec.duration))
        self.duration.textChanged.connect(self._validate)

        self.layout = QVBoxLayout()
        self.form = QFormLayout()
        self.form.addRow("Initial datetime", self.datetime)
        self.form.addRow("Initial integer", self.start_integer)
        self.form.addRow("Time step duration", self.duration)
        self.layout.addLayout(self.form)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

        self._validate()

    def _validate(self):
        try:
            Duration(self.duration.text())
        except ParameterValueFormatError:
            self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)
            return
        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(True)

    def get_spec(self):
        start_datetime = DateTime(self.datetime.dateTime().toString(Qt.ISODate))
        duration = Duration(self.duration.text())
        start_int = self.start_integer.value()
        return IntegerSequenceDateTimeConvertSpec(start_datetime, start_int, duration)


class TableViewWithButtonHeader(QTableView):
    """Customized table with data type buttons on horizontal and vertical headers"""

    def __init__(self, parent=None):
        """
        Args:
            parent (QWidget): a parent widget
        """
        super().__init__(parent)
        self._horizontal_header = HeaderWithButton(Qt.Orientation.Horizontal, self)
        self._vertical_header = HeaderWithButton(Qt.Orientation.Vertical, self)
        self.setHorizontalHeader(self._horizontal_header)
        self._horizontal_header.setContextMenuPolicy(Qt.CustomContextMenu)
        self._horizontal_header.customContextMenuRequested.connect(self._show_horizontal_header_menu)
        self._horizontal_menu = self._create_horizontal_header_menu()
        self.setVerticalHeader(self._vertical_header)
        self._vertical_header.setContextMenuPolicy(Qt.CustomContextMenu)
        self._vertical_header.customContextMenuRequested.connect(self._show_vertical_header_menu)
        self._vertical_menu = self._create_vertical_header_menu()

    def scrollContentsBy(self, dx, dy):
        """Scrolls the table's contents by given delta."""
        super().scrollContentsBy(dx, dy)
        if dx != 0:
            self._horizontal_header.fix_widget_positions()
        if dy != 0:
            self._vertical_header.fix_widget_positions()

    def _create_horizontal_header_menu(self):
        """Returns a new menu for the horizontal header"""
        parent = self._horizontal_header
        menu = QMenu(parent)
        type_menu = _create_allowed_types_menu(parent, self._set_all_column_data_types)
        type_menu.setTitle("Set all data types to...")
        menu.addMenu(type_menu)
        return menu

    def _create_vertical_header_menu(self):
        """Returns a new menu for the vertical header."""
        parent = self._horizontal_header
        menu = QMenu(parent)
        type_menu = _create_allowed_types_menu(parent, self._set_all_row_data_types)
        type_menu.setTitle("Set all data types to...")
        menu.addMenu(type_menu)
        return menu

    @Slot(QPoint)
    def _show_horizontal_header_menu(self, pos):
        """Opens the context menu of the horizontal header."""
        if self._horizontal_header.display_all or self._horizontal_header.sections_with_buttons:
            screen_pos = self._horizontal_header.mapToGlobal(pos)
            self._horizontal_menu.exec(screen_pos)

    @Slot(QPoint)
    def _show_vertical_header_menu(self, pos):
        """Opens the context menu of the vertical header."""
        if self._vertical_header.display_all or self._vertical_header.sections_with_buttons:
            screen_pos = self._vertical_header.mapToGlobal(pos)
            self._vertical_menu.exec(screen_pos)

    @Slot(QAction)
    def _set_all_column_data_types(self, action):
        """Sets all columns data types to the type given by action's text."""
        type_str = action.text()
        columns = range(self._horizontal_header.count())
        self._horizontal_header.change_data_types(columns, type_str)

    @Slot(QAction)
    def _set_all_row_data_types(self, action):
        """Sets all rows data types to the type given by action's text."""
        type_str = action.text()
        rows = range(self._vertical_header.count())
        self._vertical_header.change_data_types(rows, type_str)

    def set_undo_stack(self, undo_stack, about_to_undo_slot):
        """
        Sets undo stack for the table.

        Args:
            undo_stack (QUndoStack): undo stack
            about_to_undo_slot (Callable): slot that should be called just before the undo operation
        """
        self._horizontal_header.set_undo_stack(undo_stack)
        self._horizontal_header.about_to_undo.connect(about_to_undo_slot)
        self._vertical_header.set_undo_stack(undo_stack)
        self._vertical_header.about_to_undo.connect(about_to_undo_slot)


def _make_type_button(parent, menu, font):
    button = QToolButton(parent=parent)
    button.setMenu(menu)
    button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
    button.setFont(font)
    button.setCursor(Qt.ArrowCursor)
    return button


class HeaderWithButton(QHeaderView):
    """Class that reimplements the QHeaderView section paint event to draw a button
    that is used to display and change the type of that column or row.
    """

    about_to_undo = Signal(str)
    """Emitted when an undo/redo command is going to be executed."""

    def __init__(self, orientation, parent=None):
        """
        Args:
            orientation (Orientation): header orientation
            parent (QWidget, optional): parent widget
        """
        super().__init__(orientation, parent)
        self._source_table_name = None
        self._undo_stack = None
        self.setHighlightSections(True)
        self.setSectionsClickable(True)
        self.setDefaultAlignment(Qt.AlignLeft)
        self.sectionResized.connect(self._section_resize)
        self.sectionMoved.connect(self._section_move)
        self._font = QFont(TOOLBOX_FONT.family)

        self._display_all = True
        self._display_sections = []

        self._margin = Margin(left=0, right=0, top=0, bottom=0)

        self._menu = _create_allowed_types_menu(self, self._menu_pressed)

        self._button = _make_type_button(self, self._menu, self._font)
        self._button.hide()

        self._render_button = QToolButton(parent=self)
        self._render_button.setFont(self._font)
        self._render_button.hide()

        self.setMinimumSectionSize(self.minimumSectionSize() + self.widget_width())

    @property
    def display_all(self):
        return self._display_all

    @display_all.setter
    def display_all(self, display_all):
        self._display_all = display_all
        self.viewport().update()

    @property
    def sections_with_buttons(self):
        return self._display_sections

    @sections_with_buttons.setter
    def sections_with_buttons(self, sections):
        self._display_sections = set(sections)
        self.viewport().update()

    @Slot(QAction)
    def _menu_pressed(self, action):
        """Sets the data type of a row or column according to menu action."""
        logical_index = self.logicalIndexAt(self._button.pos())
        type_str = action.text()
        self.change_data_types(logical_index, type_str)

    def change_data_types(self, sections, type_str):
        """
        Pushes :class:`SetColumnOrRowType` to the undo stack.

        Args:
            sections (Iterable or int or NoneType): row/column index
            type_str (str): data type name
        """
        orientation = self.orientation()
        if type_str == "integer sequence datetime":
            existing_spec = None
            if sections is not None and (sections == 0 or sections):
                section = sections if isinstance(sections, int) else next(iter(sections))
                section_type = self.model().get_type(section, orientation)
                if isinstance(section_type, IntegerSequenceDateTimeConvertSpec):
                    existing_spec = section_type
            dialog = IntegerSequenceDateTimeConvertSpecDialog(existing_spec, self)
            if not dialog.exec():
                return
            convert_spec = dialog.get_spec()
        else:
            convert_spec = value_to_convert_spec(type_str)
        if not isinstance(sections, Iterable):
            sections = [sections]
        previous_convert_spec = self.model().get_type(sections[0], orientation)
        self._undo_stack.push(
            SetColumnOrRowType(self._source_table_name, self, sections, convert_spec, previous_convert_spec)
        )

    def set_data_types(self, source_table_name, sections, convert_specification):
        """
        Sets the data type for given sections.

        Args:
            source_table_name (str): name of the source table
            sections (Iterable): section indexes
            convert_specification (ConvertSpec): data conversion specification
        """
        self.about_to_undo.emit(source_table_name)
        self.model().set_types(sections, convert_specification, self.orientation())

    @Slot()
    def update_buttons(self):
        """Updates the buttons."""
        self.viewport().update()

    def widget_width(self):
        """Width of widget

        Returns:
            int: Width of widget
        """
        if self.orientation() == Qt.Orientation.Horizontal:
            return self.height()
        return self.sectionSize(0)

    def widget_height(self):
        """Height of widget

        Returns:
            int: Height of widget
        """
        if self.orientation() == Qt.Orientation.Horizontal:
            return self.height()
        return self.sectionSize(0)

    def _hide_or_show_button(self, logical_index):
        """Hides or shows the button depending on the logical index.

        Args:
            logical_index (int)
        """
        if logical_index in self._display_sections or self._display_all:
            self._set_button_geometry(self._button, logical_index)
            self._button.show()
        else:
            self._button.hide()

    def mouseMoveEvent(self, mouse_event):
        """Moves the button to the correct section so that interacting with the button works."""
        logical_index = self.logicalIndexAt(mouse_event.x(), mouse_event.y())
        self._hide_or_show_button(logical_index)
        super().mouseMoveEvent(mouse_event)

    def enterEvent(self, event):
        """Shows the button."""
        mouse_position = self.mapFromGlobal(QCursor.pos())
        logical_index = self.logicalIndexAt(mouse_position)
        self._hide_or_show_button(logical_index)
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Hides button."""
        self._button.hide()
        super().leaveEvent(event)

    def _set_button_geometry(self, button, index):
        """Sets a buttons geometry depending on the index.

        Arguments:
            button (QWidget): QWidget that geometry should be set
            index (int): logical_index to set position and geometry to.
        """
        margin = self._margin
        if self.orientation() == Qt.Orientation.Horizontal:
            button.setGeometry(
                self.sectionViewportPosition(index) + margin.left,
                margin.top,
                self.widget_width() - margin.left - margin.right,
                self.widget_height() - margin.top - margin.bottom,
            )
        else:
            button.setGeometry(
                margin.left,
                self.sectionViewportPosition(index) + margin.top,
                self.widget_width() - margin.left - margin.right,
                self.widget_height() - margin.top - margin.bottom,
            )

    def _section_resize(self, i):
        """When a section is resized.

        Arguments:
            i (int): logical index to section being resized
        """
        self._button.hide()
        mouse_position = self.mapFromGlobal(QCursor.pos())
        logical_index = self.logicalIndexAt(mouse_position)
        if i == logical_index:
            self._set_button_geometry(self._button, logical_index)

    def paintSection(self, painter, rect, logical_index):
        """Paints a section of the QHeader view.

        Works by drawing a pixmap of the button to the left of the original paint rectangle.
        Then shifts the original rect to the right so these two don't paint over each other.
        """
        if not self._display_all and logical_index not in self._display_sections:
            super().paintSection(painter, rect, logical_index)
            return

        # get the type of the section.
        type_spec = self.model().get_type(logical_index, self.orientation())
        if type_spec is None:
            type_spec = StringConvertSpec.DISPLAY_NAME
        else:
            type_spec = type_spec.DISPLAY_NAME
        font_str = TYPE_TO_FONT_AWESOME_ICON[type_spec]

        # set data for both interaction button and render button.
        self._button.setText(font_str)
        self._render_button.setText(font_str)
        self._set_button_geometry(self._render_button, logical_index)

        # get pixmap from render button and draw into header section.
        rw = self._render_button.grab()
        if self.orientation() == Qt.Orientation.Horizontal:
            painter.drawPixmap(self.sectionViewportPosition(logical_index), 0, rw)
        else:
            painter.drawPixmap(0, self.sectionViewportPosition(logical_index), rw)

        # shift rect that super class should paint in to the right so it doesn't
        # paint over the button
        rect.adjust(self.widget_width(), 0, 0, 0)
        super().paintSection(painter, rect, logical_index)

    def sectionSizeFromContents(self, logical_index):
        """Add the button width to the section so it displays right.

        Arguments:
            logical_index (int): logical index of section

        Returns:
            QSize: Size of section
        """
        org_size = super().sectionSizeFromContents(logical_index)
        org_size.setWidth(org_size.width() + self.widget_width())
        return org_size

    def _section_move(self, logical, old_visual_index, new_visual_index):
        """Section being moved.

        Arguments:
            logical (int): logical index of section beeing moved.
            old_visual_index (int): old visual index of section
            new_visual_index (int): new visual index of section
        """
        self._button.hide()
        mouse_position = self.mapFromGlobal(QCursor.pos())
        logical_index = self.logicalIndexAt(mouse_position)
        self._set_button_geometry(self._button, logical_index)

    def fix_widget_positions(self):
        """Update position of interaction button"""
        mouse_position = self.mapFromGlobal(QCursor.pos())
        logical_index = self.logicalIndexAt(mouse_position)
        self._set_button_geometry(self._button, logical_index)

    def set_margins(self, margins):
        """Sets the header margins."""
        self._margin = margins

    def setModel(self, model):
        """
        Sets the model for this view.

        Args:
            model (QAbstractItemModel): a model
        """
        if isinstance(model, SourceDataTableModel):
            old_model = self.model()
            if self.orientation() == Qt.Orientation.Horizontal:
                model.column_types_updated.connect(self.update_buttons)
                if isinstance(old_model, SourceDataTableModel):
                    old_model.column_types_updated.disconnect(self.update_buttons)
            else:
                model.row_types_updated.connect(self.update_buttons)
                if isinstance(old_model, SourceDataTableModel):
                    old_model.row_types_updated.disconnect(self.update_buttons)
        super().setModel(model)

    def set_undo_stack(self, undo_stack):
        """
        Sets undo stack for the header making menu actions to work.

        Args:
            undo_stack (QUndoStack): undo stack
        """
        self._undo_stack = undo_stack

    @Slot(QModelIndex, QModelIndex)
    def set_source_table(self, current, previous):
        """
        Sets the current source table.

        Args:
            current (QModelIndex): current source table index
            previous (QModelIndex): previous source table index
        """
        self._source_table_name = current.data()


def _create_allowed_types_menu(parent, trigger_slot):
    """
    Returns a menu which contains actions for each allowed data type.

    Args:
        parent (QWidget): a parent widget
        trigger_slot (Slot): a slot which is connected to QMenu's 'triggered' signal
    Returns:
        QMenu: a menu
    """
    menu = QMenu(parent)
    for at in _ALLOWED_TYPES:
        icon_char = TYPE_TO_FONT_AWESOME_ICON[at]
        engine = CharIconEngine(icon_char, 0)
        icon = QIcon(engine.pixmap())
        menu.addAction(icon, at)
    menu.triggered.connect(trigger_slot)
    return menu



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains settings classes for filters and manipulators."""
from spinedb_api.filters.renamer import entity_class_renamer_config, parameter_renamer_config
from spinedb_api.filters.value_transformer import value_transformer_config


class FilterSettings:
    """Base class for filter and manipulator settings."""

    def filter_config(self):
        """
        Creates filter configuration dictionary.

        Returns:
            dict: filter configuration
        """
        raise NotImplementedError()

    def report_inconsistencies(self):
        """Checks and reports errors and inconsistencies in the settings.

        Returns:
            list of str: list of warning messages
        """
        return []

    def to_dict(self):
        """
        Serializes settings to JSON compatible dictionary.

        Returns:
            dict: serialized settings
        """
        raise NotImplementedError()

    @staticmethod
    def type():
        """
        Returns a type identifier string.

        Returns:
            str: type id
        """
        raise NotImplementedError()

    @staticmethod
    def use_shorthand():
        """
        Tells if the filter config should use shorthand notation.

        Returns:
            bool: True is shorthand notation is preferred, False if config should be written on disk.
        """
        raise NotImplementedError()

    @staticmethod
    def from_dict(settings_dict):
        """
        Restores settings from dictionary.

        Args:
            settings_dict (dict): serialized settings

        Returns:
            FilterSettings: deserialized settings
        """
        raise NotImplementedError()


class RenamingSettings(FilterSettings):
    """
    Base class for renamer settings.

    Attributes:
        name_map (dict): a mapping from original name to new name
    """

    def __init__(self, name_map):
        """
        Args:
            name_map (dict): a mapping from original name to new name
        """
        self.name_map = name_map

    def __eq__(self, other):
        if not isinstance(other, RenamingSettings):
            return NotImplemented
        return self.name_map == other.name_map

    def filter_config(self):
        """See base class."""
        raise NotImplementedError()

    @staticmethod
    def type():
        """See base class."""
        raise NotImplementedError()

    @staticmethod
    def use_shorthand():
        """See base class."""
        return False

    def to_dict(self):
        """See base class."""
        return self.name_map

    @staticmethod
    def from_dict(settings_dict):
        """See base class."""
        raise NotImplementedError()


class EntityClassRenamingSettings(RenamingSettings):
    """Settings for entity class renaming manipulator."""

    def filter_config(self):
        """See base class."""
        useful = {name: rename for name, rename in self.name_map.items() if rename and name != rename}
        return entity_class_renamer_config(**useful)

    @staticmethod
    def from_dict(settings_dict):
        """See base class."""
        return EntityClassRenamingSettings(settings_dict)

    @staticmethod
    def type():
        """See base class."""
        return "entity_class_rename"


class ParameterRenamingSettings(RenamingSettings):
    """Settings for parameter renaming manipulator."""

    def filter_config(self):
        """See base class."""
        name_map = {}
        for class_name, param_renames in self.name_map.items():
            useful = {name: rename for name, rename in param_renames.items() if rename and name != rename}
            if useful:
                name_map[class_name] = useful
        return parameter_renamer_config(name_map)

    def report_inconsistencies(self):
        messages = super().report_inconsistencies()
        if any(not klass for klass in self.name_map):
            messages.append("Class name(s) missing in specification.")
        return messages

    @staticmethod
    def from_dict(settings_dict):
        """See base class."""
        if settings_dict and isinstance(next(iter(settings_dict.values())), str):
            # Legacy settings.
            settings_dict = {"": settings_dict}
        return ParameterRenamingSettings(settings_dict)

    @staticmethod
    def type():
        """See base class."""
        return "parameter_rename"


class ValueTransformSettings(FilterSettings):
    """Settings for value transformer manipulator."""

    def __init__(self, instructions):
        """
        Args:
            instructions (dict): transform instructions
        """
        self.instructions = instructions

    def filter_config(self):
        # Remove no-op instructions.
        instructions = {}
        for class_name, parameter_transformation in self.instructions.items():
            useful = {
                param_name: instructions
                for param_name, instructions in parameter_transformation.items()
                if instructions
            }
            if useful:
                instructions[class_name] = useful
        return value_transformer_config(instructions)

    def to_dict(self):
        return self.instructions

    @staticmethod
    def type():
        return "value_transformer"

    @staticmethod
    def use_shorthand():
        return False

    @staticmethod
    def from_dict(settings_dict):
        return ValueTransformSettings(settings_dict)


def settings_from_dict(settings_dict):
    """
    Restores filter settings.

    Args:
        settings_dict (dict): serialized filter settings

    Returns:
        FilterSettings: restored settings
    """
    restorers = {
        EntityClassRenamingSettings.type(): EntityClassRenamingSettings.from_dict,
        ParameterRenamingSettings.type(): ParameterRenamingSettings.from_dict,
        ValueTransformSettings.type(): ValueTransformSettings.from_dict,
    }
    return restorers.get(settings_dict["type"], lambda _: None)(settings_dict["settings"])



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ImporterIcon class."""
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtCore import QEvent
from PySide6.QtWidgets import QApplication, QGraphicsSceneMouseEvent
from spine_items.importer.importer_factory import ImporterFactory
from tests.mock_helpers import clean_up_toolbox, create_toolboxui_with_project


class TestImporterIcon(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        super().setUp()
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)
        item_dict = {"type": "Importer", "description": "", "x": 0, "y": 0, "specification": None}
        exp = ImporterFactory.make_item("I", item_dict, self._toolbox, self._toolbox.project())
        self._toolbox.project().add_item(exp)

    def tearDown(self):
        super().tearDown()
        clean_up_toolbox(self._toolbox)
        self._temp_dir.cleanup()

    def test_mouse_double_click_event(self):
        icon = self._toolbox.project()._project_items["I"].get_icon()
        with mock.patch("spine_items.importer.importer.Importer.open_import_editor") as mock_open_import_editor:
            mock_open_import_editor.return_value = True
            icon.mouseDoubleClickEvent(QGraphicsSceneMouseEvent(QEvent.Type.GraphicsSceneMouseDoubleClick))
            mock_open_import_editor.assert_called()


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`ExporterIcon`."""
from spinetoolbox.project_item_icon import ProjectItemIcon
from ..animations import AnimationSignaller, ExporterAnimation


class ExporterIcon(ProjectItemIcon):
    def __init__(self, toolbox, icon, icon_color):
        """Exporter icon for the Design View.

        Args:
            toolbox (ToolBoxUI): QMainWindow instance
            icon (str): icon resource path
            icon_color (QColor): Icon's color
        """
        super().__init__(toolbox, icon, icon_color)
        self.animation = ExporterAnimation(self)
        self.animation_signaller = AnimationSignaller()
        self.animation_signaller.animation_started.connect(self.animation.start)
        self.animation_signaller.animation_stopped.connect(self.animation.stop)

    def mouseDoubleClickEvent(self, e):
        """
        Opens transformer's specification editor.

        Args:
            e (QGraphicsSceneMouseEvent): Event
        """
        super().mouseDoubleClickEvent(e)
        item = self._toolbox.project().get_item(self._name)
        item.show_specification_window()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the custom_menus.py module."""
import logging
from pathlib import Path
import sys
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtGui import QStandardItem, QStandardItemModel
from PySide6.QtWidgets import QApplication, QWidget
from spine_items.tool.tool_specifications import ExecutableTool, JuliaTool
from spine_items.tool.widgets.custom_menus import ToolSpecificationMenu
from tests.mock_helpers import MockQSettings, create_mock_toolbox_with_mock_qsettings


class TestToolSpecificationMenu(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Overridden method. Runs once before all tests in this class."""
        try:
            cls.app = QApplication().processEvents()
        except RuntimeError:
            pass
        logging.basicConfig(
            stream=sys.stderr,
            level=logging.WARNING,
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def setUp(self):
        """Overridden method. Runs before each test."""
        self.toolbox = create_mock_toolbox_with_mock_qsettings()
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        """Overridden method. Runs after each test.
        Use this to free resources after a test if needed.
        """
        self._temp_dir.cleanup()

    def test_open_main_program_file(self):
        spec = self.make_julia_tool_spec()
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        with mock.patch("spine_items.tool.widgets.custom_menus.open_url") as mock_open_url:
            menu._open_main_program_file()
            mock_open_url.assert_called()

    def test_open_main_program_file_fails_without_path(self):
        spec = self.make_julia_tool_spec()
        spec.path = None
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        with mock.patch("spine_items.tool.widgets.custom_menus.open_url") as mock_open_url:
            menu._open_main_program_file()
            mock_open_url.assert_not_called()

    def test_open_main_program_file_fails_without_path_and_includes(self):
        spec = self.make_julia_tool_spec()
        spec.path = None
        spec.includes = None
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        with mock.patch("spine_items.tool.widgets.custom_menus.open_url") as mock_open_url:
            menu._open_main_program_file()
            mock_open_url.assert_not_called()

    def test_open_main_program_file_fails_with_bat_file(self):
        spec = self.make_exec_tool_spec()
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        with mock.patch("spine_items.tool.widgets.custom_menus.open_url") as mock_open_url:
            menu._open_main_program_file()
            mock_open_url.assert_not_called()

    def test_open_main_program_dir(self):
        spec = self.make_julia_tool_spec()
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        menu._open_main_program_dir()

    def test_open_main_program_dir_fails_without_includes(self):
        spec = self.make_julia_tool_spec()
        spec.includes = None
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        menu._open_main_program_dir()

    def test_open_main_program_dir_fails_without_path(self):
        spec = self.make_julia_tool_spec()
        spec.path = None
        spec_model = CustomQStandardItemModel()
        spec_item = QStandardItem(spec.name)
        spec_item.setData(spec)
        spec_model.appendRow(spec_item)
        index = spec_model.index(0, 0)
        w = ParentWidget(spec_model)
        menu = ToolSpecificationMenu(w, index)
        menu._open_main_program_dir()

    def make_julia_tool_spec(self):
        script_dir = Path(self._temp_dir.name, "scripts")
        script_dir.mkdir()
        script_file_name = "hello.jl"
        file_path = Path(script_dir, script_file_name)
        with open(file_path, "w") as script_file:
            script_file.writelines(["println('hello')\n"])
        mock_logger = mock.MagicMock()
        julia_tool_spec = JuliaTool(
            "test_julia_spec",
            "julia",
            str(script_dir),
            [script_file_name],
            MockQSettings(),
            mock_logger,
        )
        julia_tool_spec.init_execution_settings()  # Sets defaults
        return julia_tool_spec

    def make_exec_tool_spec(self):
        script_dir = Path(self._temp_dir.name, "scripts")
        script_dir.mkdir()
        script_file_name = "batch.bat"
        file_path = Path(script_dir, script_file_name)
        with open(file_path, "w") as script_file:
            script_file.writelines(["dir\n"])
        mock_logger = mock.MagicMock()
        return ExecutableTool(
            "test_exec_spec",
            "executable",
            str(script_dir),
            [script_file_name],
            MockQSettings(),
            mock_logger,
        )


class CustomQStandardItemModel(QStandardItemModel):
    """Fake specification model."""

    def specification(self, row):
        item = self.item(row, 0)
        return item.data()


class ParentWidget(QWidget):
    """Fake self._toolbox."""

    def __init__(self, spec_model):
        super().__init__()
        self.specification_model = spec_model
        self.msg_warning = MiniLogger()
        self.msg_error = MiniLogger()

    def open_anchor(self, url):
        """Fakes self._toolbox.open_anchor()"""
        return True


class MiniLogger:
    """Fakes calls to signal emits."""

    def emit(self, msg):
        return True



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the options_widgets.py module."""
import logging
import sys
import unittest
from PySide6.QtWidgets import QApplication, QWidget
from spine_items.tool.widgets.options_widgets import JuliaOptionsWidget


class TestJuliaOptionsWidget(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Overridden method. Runs once before all tests in this class."""
        try:
            cls.app = QApplication().processEvents()
        except RuntimeError:
            pass
        logging.basicConfig(
            stream=sys.stderr,
            level=logging.WARNING,
            format="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def test_options_widget(self):
        ow = JuliaOptionsWidget()
        ow.set_tool(QWidget())  # Obviously not a real Tool
        ow._set_ui_at_work()
        ow._set_ui_at_rest()
        options = {"julia_sysimage": "/some/path"}
        ow.do_update_options(options)
        self.assertEqual("/some/path", ow.ui.lineEdit_sysimage.text())



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Exporter's ``executable_item`` module."""
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtWidgets import QApplication
from spine_engine.project_item.project_item_resource import database_resource
from spine_items.exporter.executable_item import ExecutableItem
from spine_items.exporter.exporter import Exporter
from spine_items.exporter.output_channel import OutputChannel
from spine_items.exporter.specification import Specification
from tests.mock_helpers import clean_up_toolbox, create_toolboxui_with_project


class TestExecutableItemWithToolbox(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        """Set up."""
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)

    def tearDown(self):
        clean_up_toolbox(self._toolbox)
        self._temp_dir.cleanup()

    def test_from_dict(self):
        project = self._toolbox.project()
        specification = Specification("Test spec", "Exporter specification for testing")
        project.add_specification(specification)
        database_path = str(Path(project.project_dir, "database.sqlite"))
        output_channels = [
            OutputChannel("in label", "My exporter", "out label", {"dialect": "sqlite", "database": database_path})
        ]
        exporter = Exporter(
            "My exporter",
            "",
            0.0,
            0.0,
            self._toolbox,
            project,
            "Test spec",
            output_channels,
            output_time_stamps=False,
            cancel_on_error=True,
        )
        project.add_item(exporter)
        exporter.upstream_resources_updated(
            [database_resource("My data store", "sqlite:/// " + database_path, "in label")]
        )
        item_dict = exporter.item_dict()
        logger = mock.MagicMock()
        specifications = {exporter.item_type(): {specification.name: specification}}
        executable = ExecutableItem.from_dict(
            item_dict, "My exporter", project.project_dir, self._toolbox.qsettings(), specifications, logger=logger
        )
        self.assertEqual(executable.item_type(), exporter.item_type())
        self.assertEqual(executable.name, exporter.name)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os.path
import sys

license_text = [
    "######################################################################################################################\n",
    "# Copyright (C) 2017-2022 Spine project consortium\n",
    "# Copyright Spine Items contributors\n",
    "# This file is part of Spine Items.\n",
    "# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General\n",
    "# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)\n",
    "# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;\n",
    "# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General\n",
    "# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with\n",
    "# this program. If not, see <http://www.gnu.org/licenses/>.\n",
    "######################################################################################################################\n",
]


def append_license(path):
    _, extension = os.path.splitext(path)
    if extension not in [".py", ".ui", ".xml"]:
        raise RuntimeError("Unsupported file type. Can only append license to .py, .ui or .xml files.")
    if extension == ".py":
        _append_license_py(path)
    else:
        _append_license_xml(path)


def _append_license_py(path):
    """Appends a license header to given .py, .ui or .xml file."""
    base_name = os.path.basename(path)
    print("Appending license to " + base_name)
    with open(path) as input_file:
        contents = input_file.readlines()
    if contents[2].startswith("# Copyright"):
        print(base_name + " seems to have a license already. Skipping.")
        return
    with open(path, "w") as output_file:
        output_file.writelines(contents[:1])  # First line contains encoding.
        output_file.writelines(license_text)
        output_file.writelines(contents[1:])


def _append_license_xml(path):
    """Appends a license header to given .py, .ui or .xml file."""
    xml_license = []
    xml_license.append("<!--\n")
    for line in license_text:
        xml_license.append(line.replace("/", "\\/"))
    xml_license.append("-->\n")
    with open(path) as input_file:
        contents = input_file.readlines()
    base_name = os.path.basename(path)
    if contents[1].startswith("<!--"):
        print(base_name + " seems to have a license already. Skipping.")
        return
    print("Appending license to " + base_name)
    with open(path, "w") as output_file:
        output_file.writelines(contents[:1])  # First line contains XML schema etc.
        output_file.writelines(xml_license)
        output_file.writelines(contents[1:])


if __name__ == "__main__":
    file_name = sys.argv[1]
    append_license(file_name)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for ``output_resources`` module."""
from pathlib import PurePath
import unittest
from unittest.mock import MagicMock
from spine_engine.project_item.project_item_resource import url_resource
from spine_items.data_connection.output_resources import scan_for_resources


class TestScanForResources(unittest.TestCase):
    def test_credentials_do_not_show_in_url_resource_label(self):
        data_connection = MagicMock()
        data_connection.name = "test dc"
        project_dir_path = PurePath(__file__).parent
        data_connection.data_dir = str(project_dir_path / ".spinetoolbox" / "test_dc")
        file_paths = []
        urls = [
            {
                "dialect": "postgresql",
                "host": "long.gone.url",
                "port": 5432,
                "database": "warehouse",
                "username": "superman",
                "password": "t0p s3cr3t",
            }
        ]
        resources = scan_for_resources(data_connection, file_paths, urls, str(project_dir_path))
        self.assertEqual(
            resources,
            [
                url_resource(
                    "test dc",
                    "postgresql+psycopg2://superman:***@long.gone.url:5432/warehouse",
                    "<test dc>postgresql+psycopg2://long.gone.url:5432/warehous",
                )
            ],
        )


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Provides OptionsWidget and subclasses for each tool type (julia, python, executable, gams)."""
import os
import sys
import uuid
from PySide6.QtCore import QPointF, Qt, QVariantAnimation, Slot
from PySide6.QtGui import QBrush, QIcon, QLinearGradient, QPalette
from PySide6.QtWidgets import QFileDialog, QWidget
from spine_items.tool.utils import get_julia_path_and_project
from spine_items.utils import escape_backward_slashes
from spinetoolbox.execution_managers import QProcessExecutionManager
from spinetoolbox.helpers import CharIconEngine, get_open_file_name_in_last_dir, make_settings_dict_for_engine
from spinetoolbox.spine_engine_worker import SpineEngineWorker


class OptionsWidget(QWidget):
    def __init__(self):
        super().__init__()
        self._tool = None

    def set_tool(self, tool):
        """Init class.

        Args:
            tool (Tool)
        """
        self._tool = tool

    @property
    def _project(self):
        return self._tool.project

    @property
    def _settings(self):
        return self._project.app_settings

    @property
    def _logger(self):
        return self._tool.logger


class JuliaOptionsWidget(OptionsWidget):
    def __init__(self):
        from ..ui.julia_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__()
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self._last_sysimage_paths = {}
        self._sysimage_paths = {}
        self._sysimage_workers = {}
        self._work_animation = self._make_work_animation()
        icon_abort = QIcon(CharIconEngine("\uf057", Qt.red))
        self.ui.toolButton_abort_sysimage.setIcon(icon_abort)
        self.ui.toolButton_new_sysimage.clicked.connect(self._create_sysimage)
        self.ui.toolButton_open_sysimage.clicked.connect(self._open_sysimage)
        self.ui.toolButton_abort_sysimage.clicked.connect(self._abort_sysimage)
        self.ui.toolButton_abort_sysimage.setVisible(False)
        self.ui.lineEdit_sysimage.editingFinished.connect(self._handle_sysimage_editing_finished)

    def _make_work_animation(self):
        """
        Returns:
            QVariantAnimation
        """
        animation = QVariantAnimation()
        animation.setStartValue(0.0)
        animation.setEndValue(1.0)
        animation.setDuration(3000)
        animation.setLoopCount(-1)
        animation.valueChanged.connect(self._handle_work_animation_value_changed)
        animation.finished.connect(self._handle_work_animation_finished)
        return animation

    @property
    def sysimage_path(self):
        return self._sysimage_paths.get(self._tool)

    @sysimage_path.setter
    def sysimage_path(self, sysimage_path):
        self._sysimage_paths[self._tool] = sysimage_path

    @property
    def sysimage_worker(self):
        return self._sysimage_workers.get(self._tool)

    @sysimage_worker.setter
    def sysimage_worker(self, sysimage_worker):
        self._sysimage_workers[self._tool] = sysimage_worker

    @property
    def last_sysimage_path(self):
        return self._last_sysimage_paths.get(self._tool)

    @last_sysimage_path.setter
    def last_sysimage_path(self, last_sysimage_path):
        self._last_sysimage_paths[self._tool] = last_sysimage_path

    @property
    def sysimage_basename(self):
        return os.path.basename(self.sysimage_path)

    @Slot()
    def _handle_work_animation_finished(self):
        # pylint: disable=undefined-variable
        self.ui.lineEdit_sysimage.setPalette(qApp.palette())

    @Slot(object)
    def _handle_work_animation_value_changed(self, step):
        width = self.ui.lineEdit_sysimage.width()
        start = QPointF(-2 * width, 0)
        stop = QPointF(3 * width, 0)
        gradient = QLinearGradient(start, stop)
        # pylint: disable=undefined-variable
        highlight_color = qApp.palette().highlight().color()
        base_color = qApp.palette().base().color()
        gradient.setColorAt((step + 0) % 1, base_color)
        gradient.setColorAt((step + 0.2) % 1, base_color)
        gradient.setColorAt((step + 0.5) % 1, highlight_color)
        gradient.setColorAt((step + 0.8) % 1, highlight_color)
        gradient.setColorAt((step + 1) % 1, base_color)
        palette = QPalette()
        palette.setBrush(QPalette.Base, QBrush(gradient))
        self.ui.lineEdit_sysimage.setPalette(palette)

    def set_tool(self, tool):
        super().set_tool(tool)
        self._update_ui()

    def _update_ui(self):
        if self.sysimage_worker is not None:
            self._set_ui_at_work()
        else:
            self._set_ui_at_rest()

    def _set_ui_at_work(self):
        self._work_animation.start()
        self.ui.toolButton_abort_sysimage.setVisible(True)
        self.ui.toolButton_new_sysimage.setVisible(False)
        self.ui.toolButton_open_sysimage.setVisible(False)
        self.ui.lineEdit_sysimage.setEnabled(False)
        self.ui.lineEdit_sysimage.setText(self.sysimage_path)

    def _set_ui_at_rest(self):
        self._work_animation.stop()
        self.ui.toolButton_abort_sysimage.setVisible(False)
        self.ui.toolButton_new_sysimage.setVisible(True)
        self.ui.toolButton_open_sysimage.setVisible(True)
        self.ui.lineEdit_sysimage.setEnabled(True)
        if self.sysimage_path is None:
            self.ui.lineEdit_sysimage.setText(self.last_sysimage_path)

    def do_update_options(self, options):
        self.last_sysimage_path = options.get("julia_sysimage")
        self.ui.lineEdit_sysimage.setText(self.last_sysimage_path)

    @Slot()
    def _handle_sysimage_editing_finished(self):
        sysimage_path = self.ui.lineEdit_sysimage.text()
        self._tool.update_options({"julia_sysimage": sysimage_path})

    @Slot(bool)
    def _open_sysimage(self, _checked=False):
        """Shows a file dialog where the user can select a julia sysimage."""
        ext = "dll" if sys.platform == "win32" else "so"
        sysimage_path, _ = get_open_file_name_in_last_dir(
            self._settings,
            "openJuliaSysimage",
            self,
            "Select Julia Sysimage file",
            self._project.project_dir,
            f"Library (*.{ext})",
        )
        if not sysimage_path:
            return
        self._tool.update_options({"julia_sysimage": sysimage_path})

    @Slot(bool)
    def _abort_sysimage(self, _checked=False):
        if isinstance(self.sysimage_worker, SpineEngineWorker):
            self.sysimage_worker.stop_engine()
        elif isinstance(self.sysimage_worker, QProcessExecutionManager):
            self.sysimage_worker.stop_execution()

    def _get_sysimage_path(self):
        ext = "dll" if sys.platform == "win32" else "so"
        spec_name = self._tool.specification().name
        suggested_file_path = os.path.join(self._project.project_dir, f"{spec_name}_JuliaSysimage.{ext}")
        file_path, _ = QFileDialog.getSaveFileName(
            self, "Create Julia Sysimage file", suggested_file_path, f"Library (*.{ext})"
        )
        if not file_path:
            return None
        return file_path

    @Slot(bool)
    def _create_sysimage(self, _checked=False):
        """Creates a Julia sysimage for the specification associated with this tool.

        Executes a workflow with a modified version of this tool spec, and all its dependencies.
        See ``self._make_sysimage_spec()`` for details about the modified spec.

        The results of the workflow are used by ``self._do_create_sysimage()` to finally create the sysimage.
        """
        if self._tool.specification() is None or self._tool.specification().tooltype.lower() != "julia":
            self._logger.msg_error.emit(
                f"No Julia specification set for Tool <b>{self._tool.name}</b>. Can't create sysimage."
            )
            return
        dag = self._project.dag_with_node(self._tool.name)
        if not dag:
            return
        self.sysimage_path = self._get_sysimage_path()
        if self.sysimage_path is None:
            return
        execution_permits = {item_name: item_name == self._tool.name for item_name in dag.nodes}
        settings = make_settings_dict_for_engine(self._settings)
        settings["appSettings/makeSysImage"] = "true"  # See JuliaToolInstance.prepare()
        dag_identifier = f"containing {self._tool.name}"
        job_id = self._project.LOCAL_EXECUTION_JOB_ID
        self.sysimage_worker = self._project.create_engine_worker(
            dag, execution_permits, dag_identifier, settings, job_id
        )
        # Use the modified spec
        engine_data = self.sysimage_worker.get_engine_data()
        spec_names = {spec["name"] for type_specs in engine_data["specifications"].values() for spec in type_specs}
        spec = self._make_sysimage_spec(spec_names)
        item_dict = engine_data["items"][self._tool.name]
        item_dict["specification"] = spec.name
        options = item_dict.get("options")
        if options:
            options["julia_sysimage"] = ""  # Don't use any previous sysimages
        spec_dict = spec.to_dict()
        spec_dict["definition_file_path"] = spec.definition_file_path
        engine_data["specifications"].setdefault(self._tool.item_type(), []).append(spec_dict)
        self.sysimage_worker.set_engine_data(engine_data)
        self.sysimage_worker.finished.connect(lambda tool=self._tool: self._do_create_sysimage(tool))
        self._update_ui()
        self.sysimage_worker.start(silent=True)
        self._logger.msg_success.emit(
            f"Process to create <b>{self.sysimage_basename}</b> successfully started.\n"
            "This process might take a while, but you can keep using Spine Toolbox as normal in the meantime."
        )

    def _get_precompile_statements_filepath(self):
        return os.path.join(self._tool.data_dir, "precompile_statements_file.jl")

    def _get_loaded_modules_filepath(self):
        return os.path.join(self._tool.data_dir, "loaded_modules.txt")

    def _make_sysimage_spec(self, spec_names):
        """Returns a modified version of this tool specification that collects necessary information
        for creating the sysimage.

        This information is written into two files in the item's data dir:
            - one containing a single line with the Julia modules that are loaded by the tool
            - the other containing precompile statements collected by the option `--trace-compile`

        Returns:
            ToolSpecification
        """
        spec = self._tool.specification().clone()
        original_program_file = os.path.join(spec.path, spec.includes.pop(0))
        loaded_modules_file = self._get_loaded_modules_filepath()
        precompile_statements_file = self._get_precompile_statements_filepath()
        with open(original_program_file, "r") as original:
            original_code = original.read()
        new_code = f"""macro write_loaded_modules(ex)
    return quote
        local before = copy(Base.loaded_modules)
        local val = $(esc(ex))
        local after = copy(Base.loaded_modules)
        open("{escape_backward_slashes(loaded_modules_file)}", "w") do f
            print(f, join(setdiff(values(after), values(before)), " "))
        end
        val
    end
end
@write_loaded_modules begin 
    {original_code}
end"""
        spec.includes.insert(0, "")
        spec.cmdline_args += [f"--trace-compile={precompile_statements_file}", "-e", new_code]
        while True:
            spec.name = str(uuid.uuid4())
            if spec.name not in spec_names:
                break
        return spec

    def _do_create_sysimage(self, tool):
        """Runs when the workflow started by ``self._create_sysimage()`` has completed.

        Launches a julia process that calls ``PackageCompiler.create_sysimage()`` using the info
        collected by the referred workflow.

        Args:
            tool (Tool): The Tool that started the sysimage creation process.
                It may be different from the Tool currently using the widget.
        """
        # Replace self._tool while we run this method
        current_tool = self._tool
        self._tool = tool
        self.sysimage_worker.clean_up()
        state = self.sysimage_worker.engine_final_state()
        if state != "COMPLETED":
            user_stopped = state == "USER_STOPPED"
            errors = "\n".join(self.sysimage_worker.process_messages.get("msg_error", []))
            msg = self._make_failure_message(user_stopped, errors)
            self._logger.msg_error.emit(msg)
            self.sysimage_worker = None
            self.sysimage_path = None
            if tool is current_tool:
                self._update_ui()
            return
        loaded_modules_file = self._get_loaded_modules_filepath()
        precompile_statements_file = self._get_precompile_statements_filepath()
        _string_in_brackets = "{String}"
        with open(loaded_modules_file, "r") as f:
            modules = f.read()
        code = f"""using Pkg;
project_dir = dirname(Base.active_project());
cp(joinpath(project_dir, "Project.toml"), joinpath(project_dir, "Project.backup"); force=true);
cp(joinpath(project_dir, "Manifest.toml"), joinpath(project_dir, "Manifest.backup"); force=true);
packages = Vector{_string_in_brackets}()
try
    modules = split("{modules}", " ");
    for m in modules
        if strip(m) == ""
            continue
        end
        if any(x -> x.name == m && x.is_direct_dep, values(Pkg.dependencies())) == true
            println("Package " * m * " already installed")
        else
            try
                Pkg.add(m)
            catch e
                if isa(e, Pkg.Types.PkgError)
                    println(m * " is not a package")
                    continue
                end
            end
        end
        println("Package " * m * " installed")
        push!(packages, m)
    end
    println("Packages to add to sysimage: " * string(packages))
    Pkg.add("PackageCompiler");
    @eval import PackageCompiler
    Base.invokelatest(
        PackageCompiler.create_sysimage,
        Symbol.(packages);
        sysimage_path="{escape_backward_slashes(self.sysimage_path)}",
        project=project_dir,
        precompile_statements_file="{escape_backward_slashes(precompile_statements_file)}"
    )
finally
    cp(joinpath(project_dir, "Project.backup"), joinpath(project_dir, "Project.toml"); force=true);
    cp(joinpath(project_dir, "Manifest.backup"), joinpath(project_dir, "Manifest.toml"); force=true);
end"""
        julia, *args = get_julia_path_and_project(current_tool.specification().execution_settings, self._settings)
        args += ["-e", code]
        self.sysimage_worker = QProcessExecutionManager(self._logger, julia, args, silent=False)
        self.sysimage_worker.execution_finished.connect(lambda ret: self._handle_sysimage_process_finished(ret, tool))
        self.sysimage_worker.start_execution(workdir=self._tool.specification().path)
        # Restore the current self._tool
        self._tool = current_tool

    def _handle_sysimage_process_finished(self, ret, tool):
        """Runs when the Julia process started by ``self._do_create_sysimage()`` finishes.
        Wraps up everything.

        Args:
            ret (int): The return code of the process, 0 indicates success.
            tool (Tool): The Tool that started the sysimage creation process.
                It may be different from the Tool currently using the widget.
        """
        # Replace self._tool while we run this method
        self._tool, current_tool = tool, self._tool
        user_stopped = self.sysimage_worker.user_stopped
        error = self.sysimage_worker.process_error
        self.sysimage_worker.deleteLater()
        self.sysimage_worker = None
        if ret != 0:
            msg = self._make_failure_message(user_stopped, error)
            self._logger.msg_error.emit(msg)
            self.sysimage_path = None
        else:
            self._tool.update_options({"julia_sysimage": self.sysimage_path})
            self._logger.msg_success.emit(f"<b>{self.sysimage_basename}</b> created successfully.\n")
        if tool is current_tool:
            self._update_ui()
        # Restore the current self._tool
        self._tool = current_tool

    def _make_failure_message(self, user_stopped, errors):
        msg = f"Process to create <b>{self.sysimage_basename}</b> "
        if user_stopped:
            msg += "stopped by the user"
        else:
            msg += "failed"
            if errors:
                msg += f" with the following error(s):\n{errors}"
        return msg
